static int lonusb_dl_formware(struct usb_device *usbdev)
{
	const struct firmware *fw;
	const char *fname;
	int ret;

	switch (le16_to_cpu(usbdev->descriptor.idProduct)) 
    {
	case GESYTEC_LPU_PID:
		fname = FIRMWARE_FX;
		break;
	case GESYTEC_LPM2_PID:
		fname = FIRMWARE_FX2;
		break;
	default:
		err("Missing FW for id: %u\n", le16_to_cpu(usbdev->descriptor.idProduct));
		return -ENODEV;
	};

	ret = request_firmware(&fw, fname, &usbdev->dev);
	if (ret) 
    {
		err("Can't load firmware %s, error: %d\n", fname, ret);
		return ret;
	}

	ret = ezusb_load_firmware(usbdev, fw->data);
	release_firmware(fw);
	if (ret)
		err("Firmware download failed %d\n", ret);
	return ret;
}

static int CreateIntUrb(PDEVICE_EXTENSION pdx)
{
	int pipe;

	pdx->read_urb = usb_alloc_urb(0, GFP_KERNEL);
	if (!pdx->read_urb) 
    {
		err("No free (read) urbs available");
		return -ENOMEM;
	}
	pipe = usb_rcvintpipe(pdx->usbdev, pdx->intr_in_ep_addr);
	pdx->bulk_in_size = usb_maxpacket(pdx->usbdev, pipe, usb_pipeout(pipe));

	pdx->intr_in_buff = kmalloc(pdx->bulk_in_size, GFP_KERNEL);
	if (!pdx->intr_in_buff) 
    {
		err("Couldn't allocate intr_in_buff");
		return -ENOMEM;
	}
	usb_fill_int_urb(pdx->read_urb, pdx->usbdev, pipe,
			pdx->intr_in_buff, pdx->bulk_in_size,
			lonusb_read_int_callback, pdx,
			pdx->bInterval);
	return 0;
}

static int lonusb_probe(struct usb_interface *interface, const struct usb_device_id *id)
{
	struct usb_device *usbdev = interface_to_usbdev(interface);
	struct usb_host_interface *if_desc;
	PDEVICE_EXTENSION pdx = NULL;
	struct usb_endpoint_descriptor *endpoint;
	int pipe;
	int i;
	int ret;

    dbgz(DEBUG_OPEN_CLOSE, "%s():", __FUNCTION__);
	//pr_err("lonusb_probe(): pr_err");
	//pr_info("lonusb_probe(): pr_info");
	//pr_devel("lonusb_probe(): pr_devel");
	//dbg("lonusb_probe(): dbg");
	//err("lonusb_probe(): err");

	if (usbdev->descriptor.iProduct == 0) 
    {
		lonusb_dl_formware(usbdev);
		ret = -ENODEV;
		goto exit;
	}

	pdx = kzalloc(sizeof(DEVICE_EXTENSION), GFP_KERNEL);
	if (pdx == NULL) 
    {
		err("Out of memory");
		ret = -ENOMEM;
		goto exit;
	}

    pdx->dev = usbdev->dev;
	mutex_init(&pdx->mutex);
	pdx->usbdev = usbdev;
	pdx->interface = interface;
	pdx->remove_pending = FALSE;

    qinit(&pdx->ReadBufferQueue, maxqueuecount);
	init_waitqueue_head(&pdx->rd_wait);
	init_waitqueue_head(&pdx->wr_wait);

	/* set up the endpoint information */
	if_desc = interface->cur_altsetting;
	for (i = 0; i < if_desc->desc.bNumEndpoints; ++i) 
    {
		endpoint = &if_desc->endpoint[i].desc;
		if (usb_endpoint_is_int_in(endpoint) && !pdx->intr_in_buff) 
        {
			/* we found a interrupt in endpoint */
			pdx->intr_in_ep_addr = endpoint->bEndpointAddress;
			pdx->bInterval = endpoint->bInterval;
			ret = CreateIntUrb(pdx);
			if (ret)
				goto error;
		}

		if (usb_endpoint_is_bulk_out(endpoint) && !pdx->bulk_out_buffer) 
            {
			/* we found a bulk out endpoint */
#if REUSE_URB
            pdx->write_urb = usb_alloc_urb(0, GFP_KERNEL);
            if (!pdx->write_urb)
                {
                err("No free (write) urbs available");
                ret = -ENOMEM;
                goto error;
                }
#endif
			pipe = usb_sndbulkpipe(usbdev, endpoint->bEndpointAddress);
			pdx->bulk_out_size = usb_maxpacket(pdx->usbdev, pipe, usb_pipeout(pipe));
			pdx->bulk_out_size = USB_WRITEBUFFER_SIZE;  // use maximum needed, because usb_submit_urb() does internal divide it

			pdx->bulk_out_endpointAddr = endpoint->bEndpointAddress;
			pdx->bulk_out_buffer = kmalloc(pdx->bulk_out_size, GFP_KERNEL);
			if (!pdx->bulk_out_buffer) 
            {
				err("Couldn't allocate bulk_out_buffer");
				ret = -ENOMEM;
				goto error;
			}
		}
	}

	/* we can register the device now, as it is ready */
	usb_set_intfdata(interface, pdx);

	/* we can register the device now, as it is ready */
	ret = usb_register_dev(interface, &lonusb_class);
	if (ret) 
    {
		/* something prevented us from registering this driver */
		err("Not able to get a minor for this device.");
		usb_set_intfdata(interface, NULL);
		goto error;
	}

	ret = 0;
#if USE_PROC
	{
  #ifdef CONFIG_USB_DYNAMIC_MINORS
        // see minor_base at http://www.makelinux.net/ldd3/chp-13-sect-4  
        int minor = interface->minor;
  #else
        int minor = interface->minor - lonusb_class.minor_base;
  #endif
		sprintf(pdx->DevStats.ProcDevName, PROC_DEV_NAME "%d" /*lonusb_class.name*/, minor);
		if (HasEasylonProcDir)
        {
            //printk(KERN_DEBUG "proc_create_data(%s, %p):\n", dev->DevStats.ProcDevName, dev);
			proc_create_data(pdx->DevStats.ProcDevName, 0, NULL, &lonusb_proc_fops_dev, pdx);
        }
	}

	ProcStats.probe_cnt++ ;
#endif
    pr_info(DRIVER_DESC " now attached to " DRIVER_DEV_NAME "%d\n", pdx->minor);
 	goto exit;

error:
	lonusb_delete(pdx);
	pdx = NULL;
exit:
	return ret;
}

/**
 *      lonusb_disconnect
 *
 *      Called by the usb core when the device is removed from the system.
 */
static void lonusb_disconnect(struct usb_interface *interface)
{
	PDEVICE_EXTENSION pdx = usb_get_intfdata(interface);
    dbgz(DEBUG_OPEN_CLOSE, "%s():", __FUNCTION__);

	usb_set_intfdata(interface, NULL);

	/* give back our minor */
	usb_deregister_dev(interface, &lonusb_class);

	mutex_lock(&pdx->mutex);

	pdx->remove_pending = TRUE;
	wake_up_interruptible(&pdx->rd_wait);
	wake_up_interruptible(&pdx->wr_wait);

	/* if the device is not opened, then we clean up right now */
	if (!pdx->open_count) 
    {
		mutex_unlock(&pdx->mutex);
		lonusb_delete(pdx);
	}
	else 
    {
		/* shutdown any bulk writes that might be going on */
		usb_kill_urb(pdx->read_urb);
#if REUSE_URB
        usb_kill_urb (pdx->write_urb);
#endif
		pdx->usbdev = NULL;
		mutex_unlock(&pdx->mutex);
	}
#if USE_PROC
	if (HasEasylonProcDir)
		remove_proc_entry(pdx->DevStats.ProcDevName, NULL);

	ProcStats.disconnect_cnt++ ;
#endif

    pr_info(DRIVER_DESC " " DRIVER_DEV_NAME "%d now disconnected\n", pdx->minor);
}

static int __init lonusb_init(void)
{
    int result;

#if USE_PROC
    if (CreateEasylonProcDir())
        proc_create_data(PROC_DEV_NAME, 0, NULL, &lonusb_proc_fops_all, NULL);
#endif

    result = usb_register(&lonusb_driver);
    if (result < 0) 
		{
        err("usb_register failed for the " DRIVER_MOD_NAME " driver. Error number %d", result);
        return -1;
        }

    pr_info(DRIVER_DESC " " DRIVER_VERSION " result=%d\n", result);
    return 0;
    }

static void __exit lonusb_cleanup(void)
{
#if USE_PROC
	if (HasEasylonProcDir)
		remove_proc_entry(PROC_DEV_NAME, NULL);
    RemoveEasylonProcDir();
#endif
    usb_deregister(&lonusb_driver);
}
module_exit(lonusb_cleanup);
module_init(lonusb_init);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");

MODULE_INFO(supported, "external");

