/*
 * Linux driver for Easylon USB Interfaces
 *
 * Copyright (c)2003 Gesytec GmbH, Volker Schober (info@gesytec.de)
 *
 */

#include "lonusb.h"

#include "dynbuf.c"

#include "driverdata.c"
#include "proc.c"
#include "ezusb.c"
#include "driverentry.c"
#include "readwrite.c"


static void DumpBuffer(int DebugZone, const char *Prefix, int size, const unsigned char *data)
    {
    if (debug & DebugZone)
		print_hex_dump(KERN_DEBUG, Prefix, DUMP_PREFIX_OFFSET, 32, 1, data, size, false);
    }

static void lonusb_delete(PDEVICE_EXTENSION pdx)
{
	kfree(pdx->intr_in_buff);
	kfree(pdx->bulk_out_buffer);
#if REUSE_URB
    if (pdx->write_urb != NULL)
        {
		dbg("usb_free_urb(dev->write_urb)");
        usb_free_urb (pdx->write_urb);
        }
#endif
	if (pdx->read_urb != NULL) 
    {
		dbg("usb_free_urb(dev->read_urb)");
		usb_free_urb(pdx->read_urb);
	}

	pdx->RemainDataSize = 0;
    qexit(&pdx->ReadBufferQueue);
    kfree(pdx);
}

static int lonusb_open(struct inode *inode, struct file *file)
{
	int retval = 0;
	PDEVICE_EXTENSION pdx = NULL;
	int subminor = iminor(inode);

	dbgz(DEBUG_OPEN_CLOSE, "%s(): subminor=%d", __FUNCTION__, subminor);

    struct usb_interface *interface = usb_find_interface(&lonusb_driver, subminor);

	if (!interface) 
    {
		err("error, can't find device for minor %d", subminor);
		retval = -ENODEV;
		goto exit;
	}

	pdx = usb_get_intfdata(interface);
	if (!pdx) 
    {
		retval = -ENODEV;
		goto exit;
	}

    {
        dbg("lonusb_open(subminor=%d) minor_base=%d", subminor, lonusb_class.minor_base);
    }

	pdx->minor = subminor;

	mutex_lock(&pdx->mutex);

	if ((++pdx->open_count) > 1) 
    {
		retval = -EBUSY;
		goto error;
	}

	file->private_data = pdx;

	pdx->RemainDataSize = 0;

	retval = usb_submit_urb(pdx->read_urb, GFP_KERNEL);
    pdx->DevStats.open++;

error:
	if (retval < 0)
		--pdx->open_count;

	mutex_unlock(&pdx->mutex);
exit:
	return retval;
}

static int lonusb_release(struct inode *inode, struct file *file)
{
	PDEVICE_EXTENSION pdx = file->private_data;
	int retval = 0;

	if (pdx == NULL)
    {
        err("%s() object is NULL", __FUNCTION__);
		return -ENODEV;
    }

    dbgz(DEBUG_OPEN_CLOSE, "%s(): subminor=%d open_count=%d", __FUNCTION__, iminor(inode), pdx->open_count);

    mutex_lock(&pdx->mutex);
	if (pdx->usbdev == NULL) 
    {
		/* the device was unplugged before the file was released */
		mutex_unlock(&pdx->mutex);
		lonusb_delete(pdx);
		return 0;
	}

	if (--pdx->open_count <= 0) 
    {
		/* shutdown any bulk writes that might be going on */
		usb_kill_urb(pdx->read_urb);

#if REUSE_URB
        usb_kill_urb (pdx->write_urb);
#endif
		pdx->RemainDataSize = 0;
        qflush(&pdx->ReadBufferQueue);
		pdx->open_count = 0;
	}

	mutex_unlock(&pdx->mutex);
    pdx->DevStats.close++;
	return retval;
}
