/*
 * Linux driver for Easylon USB Interfaces
 * Copyright (c)2003 Gesytec GmbH, Volker Schober (info@gesytec.de)
 *
 * based on the generic LPC/LPP driver written by Ralf Feiland
 *
 * Minor number  0..31 are used for ISA cards
 * Minor number 32..63 are used for PCI cards
 *
 * The I/O addresses and IRQs of ISA cards can now be set at load time, so
 * it is not necessary to change the driver code.
 * ISA parameters are defined here as 16 bit values (4 hex digits)
 *   Bits  11..0:  base I/O address
 *   Bits 15..12:  IRQ, valid values are 3,5,7,9,A,B,C and F.
 *
 * The parameter isa is a string with 4 hex digits for each card.
 * All characters except 0..9, A..F and a..f are ignored and may be used
 * as seperators.
 * Also some other characters like comma and double quote are not allowed
 * by the insmod command.
 *
 * Example:
 * 1st card has address 0x340 and IRQ 5
 * 2nd card has address 0x344 and IRQ 7
 *
 * insmod lpclinux.o isa="53407344"
 * or
 * insmod lpclinux.o isa="5/340 7/344"
 * or
 * insmod lpclinux.o isa="irq=5 port=340h : irq=7 port=344h"
 *
 * Not allowed is
 * insmod lpclinux.o isa="interrupt = 5, baseaddress = 0x340"
 *                            ^         ^ ^^ ^^   ^     ^
 */

#include "lpclinux.h"

static int major;
static char *isa;

/* Module paramaters */
#define DEBUG_DUMP1	0x0001
#define DEBUG_DUMP2	0x0002
static int debug;
module_param(debug, int,  S_IRUGO | S_IWUSR);
module_param(isa, charp,  S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(debug, "Debug enabled or not");

static struct lpc_dev *minor_table[MAX_DEVICES];

/* lock to protect the minor_table structure */
static DEFINE_MUTEX(minor_table_mutex);

static struct class *lpcclass;

static const u8 irq_table[16] = {
	/* 0     1     2   *3*    4    *5*    6    *7* */
	0x00, 0x00, 0x00, 0x80, 0x00, 0x90, 0x00, 0xA0,
	/* 8   *9*  *10*  *11*  *12*    13    14  *15* */
	0x00, 0xB0, 0xC0, 0xD0, 0xE0, 0x00, 0x00, 0xF0
};

static u16 lpcs[MAX_LPCS];/* = { 0x5340, 0x7344 }; */

//static DEFINE_PCI_DEVICE_TABLE(lpp_pci_tbl) = {
static const struct pci_device_id lpp_pci_tbl[] = {
	{ PCI_DEVICE(GESYTEC_PCI_VID, GESYTEC_LPP_DID) },
	{ }
};
MODULE_DEVICE_TABLE(pci, lpp_pci_tbl);

static int check_hs(struct lpc_dev *dev)
{
	u8 st;

	while ((st = inb(dev->cport)) & 1)
		if (!(st & 4))
			return -EINVAL;
	return 0;
}

static int check_wtchs(struct lpc_dev *dev, unsigned long t)
{
	while (inb(dev->cport) & 2) {
		if (time_after(jiffies, t)) {
			if (inb(dev->cport) & 2)
				return -ENODEV;
			else
				return 0;
		}
		cpu_relax();
	}
	return 0;
}

static void DumpBuffer(char *prefix, void *pvBuffer, unsigned int length)
{
	pr_notice("%s\n", prefix);
	print_hex_dump(KERN_CONT, "", DUMP_PREFIX_OFFSET,
			16, 1,
			pvBuffer, length, false);
}


static int downlink(struct lpc_dev *dev)
{
	u8 cmq, len;
	u8 *p;
	int idx;
	u32 wport;

	wport = dev->wport;
	idx = dev->outhead;
	p = dev->outbuf[idx];
	cmq = *p++;
	len = *p++;
	if (debug & DEBUG_DUMP2)
		DumpBuffer(">>", p-2, len+2);
	if (check_hs(dev))
		return -1;
	outb(CMD_XFER, wport);
	if (check_hs(dev))
		return -1;
	outb(len + 1, wport);
	if (check_hs(dev))
		return -1;
	outb(cmq, wport);
	if (check_hs(dev))
		return -1;

	while (len--) {
		if (check_hs(dev))
			return -1;
		outb(*p++, wport);
	}
	if (check_hs(dev))
		return -1;
	outb(EOM, wport);
	idx++;
	if (idx >= NUM_OUTBUF)
		idx = 0;
	dev->outhead = idx;
	return 0;
}

static int uplink(struct lpc_dev *dev)
{
	u8 cmq, len;
	u8 *p;
	int idx;
	u32 cport;
	u32 rport;

	cport = dev->cport;
	rport = dev->rport;
	if (check_hs(dev))
		return -1;
	if (!inb(rport))
		/* Pass token */
		return 0;
	if (check_hs(dev))
		return -1;
	len = inb(rport);
	if (check_hs(dev))
		return -1;
	cmq = inb(rport);
	if (cmq == 0xC0)
		/* niACK received */
		return 1;

	/* uplink */
	idx = dev->intail;
	p = dev->inbuf[idx];
	*p++ = cmq;
	len--;
	*p++ = len;
	while (len--) {
		if (check_hs(dev))
			return -1;
		*p++ = inb(rport);
	}
	if (debug & DEBUG_DUMP2) {
		p    = dev->inbuf[idx];
		len  = p[1];
		DumpBuffer("<<", p, len+2);
	}
	idx++;
	if (idx >= NUM_INBUF)
		idx = 0;
	if (idx != dev->inhead) {
		dev->intail = idx;
		if (waitqueue_active(&dev->rd_wait))
			wake_up_interruptible(&dev->rd_wait);
	}
	return 0;
}


static int sm_lpc(struct lpc_dev *dev)
{
	u8  *p;
	int g;
	u16 idx;
	u32 cport, wport, rport;

	cport = dev->cport;
	wport = dev->wport;
	rport = dev->rport;

	switch (dev->state) {
	case ST_RESYNC:
		if (inb(cport) & 1)
			return 0;
		dev->ctl &= ~4;
		/* clear reset flipflop */
		outb(dev->ctl, cport);
		outb(dev->ctl, cport);
		outb(dev->ctl, cport);
		dev->ctl |= 4;
		outb(dev->ctl, cport);
		if (check_hs(dev))
			return -1;
		outb(CMD_RESYNC, wport);
		if (check_hs(dev))
			return -1;
		outb(EOM, wport);
		if (check_hs(dev))
			return -1;
		if (inb(rport) != CMD_ACKSYNC)
			/* get ACKSYNC */
			return -1;
		dev->state = ST_RESET;
		return 0;

	case ST_RESET:
		if (check_hs(dev))
			return -1;
		outb(CMD_NULL, wport);
		if (check_hs(dev))
			return -1;
		outb(EOM, wport);
		dev->state = ST_FLCAN;
		return 0;

	case ST_FLCAN:
		if (check_hs(dev))
			return -1;
		if (inb(rport) != CMD_XFER) {
			/* get niRESET */
			return -1;
		}
		if (check_hs(dev))
			return -1;
		if (inb(rport) != 1)
			return -1;
		if (check_hs(dev))
			return -1;
		if (inb(rport) != 0x50)
			return -1;

		idx = dev->intail;
		p = dev->inbuf[idx];
		*p++ = 0x50;
		*p++ = 0x00;
		idx++;
		if (idx >= NUM_INBUF)
			idx = 0;
		if (idx != dev->inhead) {
			dev->intail = idx;
			if (waitqueue_active(&dev->rd_wait))
				wake_up_interruptible(&dev->rd_wait);
		}
		if (check_hs(dev))
			return -1;
		/* send FLUSH_CANCEL */
		outb(CMD_XFER, wport);
		if (check_hs(dev))
			return -1;
		outb(1, wport);
		if (check_hs(dev))
			return -1;
		outb(0x60, wport);
		if (check_hs(dev))
			return -1;
		outb(EOM, wport);
		dev->state = ST_UPLINK;
		return 0;

	case ST_IDLE:
		if (dev->outhead == dev->outtail) {
			if (check_hs(dev))
				return -1;
			/* send NULL */
			outb(CMD_NULL, wport);
			if (check_hs(dev))
				return -1;
			outb(EOM, wport);
			dev->state = ST_UPLINK;
			return 0;
		}
		p = dev->outbuf[dev->outhead];
		if (!((p[0] - 0x10) & 0xE0)) {
			/* 1x or 2x */
			if (check_hs(dev))
				return -1;
			/* buffer request */
			outb(CMD_XFER, wport);
			if (check_hs(dev))
				return -1;
			outb(0x01, wport);
			if (check_hs(dev))
				return -1;
			outb(p[0], wport);
			if (check_hs(dev))
				return -1;
			outb(EOM, wport);
			if (check_hs(dev))
				return -1;
			dev->state = ST_NIACK;
			return 0;
		}
		if (downlink(dev))
			return -1;
		dev->state = ST_UPLINK;
		return 0;

	case ST_UPLINK:
		if (uplink(dev) < 0)
			return -1;
		dev->state = ST_IDLE;
		if (dev->outhead != dev->outtail)
			dev->active++;
		return 0;

	case ST_NIACK:
		g = uplink(dev);
		if (g < 0)
			return -1;
		if (g == 1) {
			downlink(dev);
			dev->state = ST_UPLINK;
			return 0;
		}
		if (check_hs(dev))
			return -1;
		/* send NULL */
		outb(CMD_NULL, wport);
		if (check_hs(dev))
			return -1;
		outb(EOM, wport);
		dev->state = ST_NIACK;
		return 0;
	}
	return 0;
}

static bool sm_stub(struct lpc_dev *dev)
{
	int result;
	u32 cport;

	if (++dev->active > 1) {
		dev_dbg(dev->device, "active=%X", dev->active);
		return false;
	}
	do  {
		result = sm_lpc(dev);
		if (result < 0) {
			/* will be 0 */
			dev->active = 1;
			dev->state = 0;
			cport = dev->cport;
			dev->ctl |= 0x01;
			outb(dev->ctl, cport);
			outb(dev->ctl, cport);
			dev->ctl &= 0xFE;
			outb(dev->ctl, cport);
		}
	} while (--dev->active > 0);
	mod_timer(&dev->drvtimer, jiffies + msecs_to_jiffies(200));
	return true;
}

static irqreturn_t lpc_isa_interrupt(int irq, void *ptr)
{
	struct lpc_dev *dev = ptr;

	if (inb(dev->cport) & 8)
		return IRQ_NONE;

	del_timer(&dev->drvtimer);
	/* Clear interrupt flipflop (ISA) */
	outb(0, dev->rport);

	spin_lock(&dev->lock);
	sm_stub(dev);
	spin_unlock(&dev->lock);

	return IRQ_HANDLED;
}

static irqreturn_t lpc_pci_interrupt(int irq, void *ptr)
{
	struct lpc_dev *dev = ptr;
	u16 IntCsr;

	IntCsr = inw(dev->PlxIntCsrPort);
	if (IntCsr == 0xFFFF)
		/* Can't read PLX */
		return IRQ_NONE;
	if (inb(dev->cport) & 8)
		return IRQ_NONE;

	del_timer(&dev->drvtimer);

	spin_lock(&dev->lock);
	/* Clear interrupt flipflop (PCI) */
	dev->ctl &= 0xEF;
	outb(dev->ctl, dev->cport);
	outb(dev->ctl, dev->cport);
	outb(dev->ctl, dev->cport);

	/* CLR_IFF_CMND */
	dev->ctl |= 0x10;
	outb(dev->ctl, dev->cport);

	sm_stub(dev);
	spin_unlock(&dev->lock);
	return IRQ_HANDLED;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
static void lpc_timer(unsigned long ptr)
{
	struct lpc_dev *dev = (struct lpc_dev *)ptr;
#else
void lpc_timer(struct timer_list *t)
{
    struct lpc_dev *dev = from_timer(dev, t, drvtimer);
#endif	

	unsigned long flags;
	spin_lock_irqsave(&dev->lock, flags);
	sm_stub(dev);
	spin_unlock_irqrestore(&dev->lock, flags);
}

static ssize_t lpc_write(struct file *file, const char __user *buf,
		size_t count, loff_t *off)
{
	struct lpc_dev *dev = file->private_data;
	u8 *p;
	int idx;
	int len;
	int ret;

	if (count < 0)
		return -EINVAL;
	if (count > 256)
		return -EINVAL;

	mutex_lock(&dev->mutex);
	idx = dev->outtail + 1;
	if (idx >= NUM_OUTBUF)
		idx = 0;

	if (idx == dev->outhead) {
		ret = -EWOULDBLOCK;
		goto err;
	}

	p = dev->outbuf[dev->outtail];
	if (copy_from_user(p, buf, count)) {
		ret = -EFAULT;
		goto err;
	}

	if (debug & DEBUG_DUMP1)
		DumpBuffer("> ", p, count);
	len = p[1] + 2;
	dev->outtail = idx;

	spin_lock_irq(&dev->lock);
	if (dev->state == ST_IDLE) {
		del_timer(&dev->drvtimer);
		sm_stub(dev);
	}
	spin_unlock_irq(&dev->lock);
	mutex_unlock(&dev->mutex);
	return len;
err:
	mutex_unlock(&dev->mutex);
	return ret;
}

static ssize_t lpc_read(struct file *file, char __user *buf,
		size_t count, loff_t *off)
{
	struct lpc_dev *dev = file->private_data;
	int idx;
	int len;
	int ret;
	u8 *p;

	if (count < 0)
		return -EINVAL;

	mutex_lock(&dev->mutex);
	idx = dev->inhead;
	if (idx == dev->intail) {
		if (file->f_flags & O_NONBLOCK) {
			ret = -EAGAIN;
			goto err;
		}

		if (wait_event_interruptible(dev->rd_wait,
					(idx != dev->intail))) {
			ret = -ERESTARTSYS;
			goto err;
		}
	}

	p = dev->inbuf[idx];

	len = p[1] + 2;
	if (count > len)
		count = len;

	if (debug & DEBUG_DUMP1)
		DumpBuffer("< ", p, count);
	if (copy_to_user(buf, p, count)) {
		ret = -EFAULT;
		goto err;
	}

	idx++;
	if (idx >= NUM_INBUF)
		idx = 0;
	dev->inhead = idx;

	mutex_unlock(&dev->mutex);
	return count;
err:
	mutex_unlock(&dev->mutex);
	return ret;
}

static unsigned lpc_poll(struct file *file, struct poll_table_struct *wait)
{
	struct lpc_dev *dev = file->private_data;
	unsigned mask = 0;

	poll_wait(file, &dev->rd_wait, wait);
	if (dev->inhead != dev->intail)
		mask |= POLLIN | POLLRDNORM;
	return mask;
}

static int initialize_isacard(struct lpc_dev *dev)
{
	u32 cport;
	unsigned int irq;

	cport = dev->cport;
	irq   = dev->irq;

	dev->ctl = 0x0F;
	outb(dev->ctl, cport);
	outb(dev->ctl, cport);
	dev->ctl = 0x0D;
	outb(dev->ctl, cport);
	outb(dev->ctl, cport);
	dev->ctl |= irq_table[irq];
	outb(dev->ctl, cport);
	outb(dev->ctl, cport);
	dev->ctl &= 0xF6;
	outb(dev->ctl, cport);

	if (request_irq(irq, lpc_isa_interrupt, IRQF_SHARED, DRIVER_DEV_NAME,
				dev)) {
		pr_err("unable to register IRQ%u\n", irq);
		return -ENODEV;
	}
	return 0;
}

static int initialize_pcicard(struct lpc_dev *dev)
{
	u32  cport;
	u16  IntCsr;
	int ret;

	cport = dev->cport;

	dev->ctl = 0x0D;
	outb(dev->ctl, cport);
	outb(dev->ctl, cport);

	dev->ctl = 0x16;
	outb(dev->ctl, cport);
	outb(dev->ctl, cport);

	dev->ctl = 0x02;
	outb(dev->ctl, cport);
	outb(dev->ctl, cport);

	dev->ctl = 0x16;
	outb(dev->ctl, cport);

	IntCsr = inw(dev->PlxIntCsrPort);
	IntCsr |= PLX_9050_INTR_ENABLE | PLX_9050_LINT1_ENABLE;
	outw(IntCsr, dev->PlxIntCsrPort);

	ret = request_irq(dev->irq, lpc_pci_interrupt, IRQF_SHARED,
			DRIVER_DEV_NAME, dev);
	if (ret) {
		pr_err("unable to register IRQ%d\n", dev->irq);
		return -ENODEV;
	}
	return 0;
}

static int lpc_release(struct inode *inode, struct file *file)
{
	struct lpc_dev *dev = file->private_data;

	if (dev == NULL)
		return -ENODEV;

	mutex_lock(&dev->mutex);
	if (dev->open_count == 0) {
		mutex_unlock(&dev->mutex);
		return 0;
	}
	--dev->open_count;

	dev->ctl = 0x0D;
	outb(dev->ctl, dev->cport);

	if (dev->is_isa) {
		outw(0, dev->rport);
	} else {
		u16 IntCsr;

		IntCsr = inw(dev->PlxIntCsrPort);
		IntCsr &= ~PLX_9050_INTR_ENABLE;
		outw(IntCsr, dev->PlxIntCsrPort);
	}

	dev->active = 1;
	free_irq(dev->irq, dev);
	del_timer_sync(&dev->drvtimer);

	dev->open_count = 0;
	mutex_unlock(&dev->mutex);
	return 0;
}

static int lpc_open(struct inode *inode, struct file *file)
{
	struct lpc_dev *dev;
	int minor;
	int idx;
	int ret;

	minor = iminor(inode);
	if (minor > MAX_DEVICES)
		return -ENODEV;

	dev = minor_table[minor];
	if (!dev) {
		pr_err("%s - error, can't find device for minor %d\n", __func__,
				minor);
		return -ENODEV;
	}

	mutex_lock(&dev->mutex);
	if (dev->open_count > 0) {
		ret = -EBUSY;
		goto out;
	}

	file->private_data = dev;
	dev->open_count++;
	dev->active = 1;

	if (dev->is_isa)
		ret = initialize_isacard(dev);
	else
		ret = initialize_pcicard(dev);
	if (ret) {
		ret = -ENODEV;
		goto err;
	}

	spin_lock_irq(&dev->lock);

	dev->inhead  = dev->intail = 0;
	dev->outhead = dev->outtail = 0;
	dev->state   = ST_RESYNC;
	dev->active  = 0;
	dev->wtc_exist = 0;

	/* Start state machine */
	sm_stub(dev);
	spin_unlock_irq(&dev->lock);

	idx = dev->inhead;
	if (idx == dev->intail) {
		if (wait_event_interruptible(dev->rd_wait,
					(idx != dev->intail))) {
			ret = -ERESTARTSYS;
			mutex_unlock(&dev->mutex);
			lpc_release(inode, file);
			return ret;
		}
	}
	idx++;
	if (idx >= NUM_INBUF)
		idx = 0;
	dev->inhead = idx;
out:
	mutex_unlock(&dev->mutex);
	return ret;
err:
	dev->open_count--;
	goto out;
}

static int lpc_watcher(struct lpc_dev *dev, void __user *arg)
{
	u8 __user *s;
	u8 __user *d;
	u8  xcvr, cmd, tmp;
	s16 count, i;
	u32 cport, wtcport;
	unsigned long t;
	int ret;

	cport   = dev->cport;
	wtcport = dev->wtcport;

	s = arg;
	d = arg;

	/* Get the length */
	ret = get_user(count, s);
	if (ret)
		return ret;
	s += sizeof(count);
	ret = get_user(cmd, s++);
	if (ret)
		return ret;
	cmd &= 0x7F;

	if (cmd) {
		/* Normal case: request watcher firmware */
		if (!dev->wtc_exist)
			return -ENODEV;

		t = jiffies + jiffies_to_msecs(500);
		if (check_wtchs(dev, t))
			return -ENODEV;

		/* Send the length */
		outb((u8)(count >> 8), wtcport);
		if (check_wtchs(dev, t))
			return -ENODEV;

		outb((u8)(count & 0xFF), wtcport);
		if (check_wtchs(dev, t))
			return -ENODEV;

		outb(cmd, wtcport);
		count--;
		while (count > 0) {
			if (check_wtchs(dev, t))
				return -ENODEV;

			ret = get_user(tmp, s++);
			if (ret)
				return ret;
			/* Send the data bytes */
			outb(tmp, wtcport);
			count--;
		}
		if (check_wtchs(dev, t))
			return -ENODEV;

		/* EOM */
		outb(0, wtcport);

		if (check_wtchs(dev, t))
			return -ENODEV;

		/* Receive the length */
		count = inb(wtcport);
		ret = put_user(count & 0xFF, d++);
		if (ret)
			return ret;
		if (check_wtchs(dev, t))
			return -ENODEV;

		count = (count << 8) + inb(wtcport);
		ret = put_user(count & 0xFF, d++);
		if (ret)
			return ret;

		i = count;
		while (i > 0) {
			if (check_wtchs(dev, t))
				return -ENODEV;

			tmp = inb(wtcport);
			ret = put_user(tmp, d++);
			if (ret)
				return ret;
			i--;
		}
	} else {
		/* Special case: firmware download */

		/* Reset LWA */
		dev->ctl |= 0x02;
		outb(dev->ctl, cport);
		/* Switch to Bootloader */
		inb(wtcport);
		/* Start LWA */
		dev->ctl &= ~0x02;
		outb(dev->ctl, cport);
		ret = get_user(xcvr, s++);
		if (ret)
			return ret;

		count -= 2;
		schedule();
		t = jiffies + msecs_to_jiffies(1000);

		while (count > 0) {
			if (check_wtchs(dev, t))
				return -ENODEV;

			get_user(tmp, s++);
			/* Send the data bytes */
			outb(tmp, wtcport);
			count--;
		}
		if (check_wtchs(dev, t))
			return -ENODEV;

		/* EOM */
		outb(0, wtcport);
		/* Wait until finished */
		if (check_wtchs(dev, t))
			return -ENODEV;

		dev->ctl |= 0x02;
		/* Reset LWA */
		outb(dev->ctl, cport);
		/* Switch to Firmware */
		outb(xcvr, wtcport);
		dev->ctl &= ~0x02;
		/* Start LWA */
		outb(dev->ctl, cport);

		/* Give the LWA some time to answer */
		t = msecs_to_jiffies(2000);
		do {
			t--;
			if (!t)
				return -ENODEV;
			schedule();
		} while (inb(cport) & 2);
		ret = put_user(0, d++);
		if (ret)
			return ret;
		ret = put_user(0, d++);
		if (ret)
			return ret;
		dev->wtc_exist = 1;
	}
	return d - (u8 __user *)arg;
}

static long lpc_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	struct lpc_dev *dev = file->private_data;
	int retval;

	switch (cmd) {
    char tmpBuf[80];
    case IOCTL_VERSION:
        retval = sprintf(tmpBuf, DRIVER_DESC " " DRIVER_VERSION ", Linux %d.%d.%d",
            (LINUX_VERSION_CODE >> 16) & 0xFF,
            (LINUX_VERSION_CODE >> 8) & 0xFF,
            LINUX_VERSION_CODE & 0xFF ) + 1;
        if (copy_to_user ((char *)arg, tmpBuf, retval))
            retval = -EFAULT;
        break;     
	case IOCTL_WATCHER:
		return lpc_watcher(dev, (void __user *)arg);
	default:
		retval = -EINVAL;
	}
	return retval;
}

static void cleanup_isa(void)
{
	int i;

	mutex_lock(&minor_table_mutex);
	for (i = 0; i < MAX_LPCS; i++) {
		if (lpcs[i])
			release_region(lpcs[i] & 0x0FFF, 4);
		if (minor_table[i]) {
			device_destroy(lpcclass, MKDEV(major, i));
			kfree(minor_table[i]);
		}
	}
	mutex_unlock(&minor_table_mutex);
}

static void common_init(struct lpc_dev *dev)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
	init_timer(&dev->drvtimer);
	dev->drvtimer.function = lpc_timer;
	dev->drvtimer.data = (unsigned long)dev;
#else
	timer_setup(&dev->drvtimer, lpc_timer, 0);
	/* the third argument may include TIMER_* flags */
#endif
	spin_lock_init(&dev->lock);
	mutex_init(&dev->mutex);
	init_waitqueue_head(&dev->rd_wait);
}

static int lpc_init(void)
{
	int i;
	u32 iobase, cport, rport;
	int irq;
	int rc = 0;

	for (i = 0; i < MAX_LPCS; i++) {
		int hwfail = 0;
		struct lpc_dev *dev;
		uint minor;

		if (!lpcs[i])
			continue;

		iobase = lpcs[i] & 0x0FFF;
		irq    = lpcs[i] >> 12;
		cport = iobase + 1;
		rport = iobase + 2;

		if (((u8)irq > 15) || !irq_table[irq]) {
			pr_err("IRQ%d not supported\n", irq);
			lpcs[i] = 0;
			continue;
		}
		if (!request_region(iobase, 4, DRIVER_DEV_NAME)) {
			pr_err("unable to register port 0x%x", iobase);
			lpcs[i] = 0;
			continue;
		}

		outb(0x5F, cport);
		outb(0, rport);
		outb(0, rport);
		outb(0, rport);

		if ((inb(cport) & 0xF0) != 0x50)
			hwfail = 1;

		outb(0x0F, cport);
		outb(0, rport);
		outb(0, rport);
		outb(0, rport);
		if (inb(cport) & 0xF0)
			hwfail = 1;
		if (hwfail) {
			pr_err("interface card not responding at port 0x%04X",
					iobase);
			lpcs[i] = 0;
			release_region(iobase, 4);
			continue;
		}

		/* select a "subminor" number (part of a minor number) */
		mutex_lock(&minor_table_mutex);
		for (minor = 0; minor < MAX_LPCS; ++minor) {
			if (minor_table[minor] == NULL)
				break;
		}
		if (minor >= MAX_LPCS) {
			pr_info("Too many devices plugged in, can not " \
					"handle this device.");
			rc = -EINVAL;
			goto err_minor_table;
		}

		dev = kzalloc(sizeof(struct lpc_dev), GFP_KERNEL);
		if (!dev) {
			rc = -ENOMEM;
			goto err_minor_table;
		}

		common_init(dev);

		dev->is_isa = true;
		dev->cport   = iobase + 1;
		dev->rport   = iobase + 2;
		dev->wport   = iobase + 3;
		dev->wtcport = iobase + 0;
		dev->irq     = irq;
		dev->PlxIntCsrPort = 0;

		dev->minor = minor;
		minor_table[minor] = dev;
		mutex_unlock(&minor_table_mutex);

		dev->device = device_create(lpcclass, NULL, MKDEV(major, minor),
				NULL, "lpcdrv%u", minor);
		if (!dev->device) {
			pr_err("device_create failed for lpcdrv\n");
			kfree(dev);
			rc = -EINVAL;
			goto err_minor_table;
		}
	}
	return rc;

err_minor_table:
	mutex_unlock(&minor_table_mutex);
	release_region(iobase, 4);
	cleanup_isa();
	return rc;
}

#ifndef CONFIG_PCI
static int pci_set_mwi(struct pci_dev *pdev)
{
	return 0;
}

static void pci_clear_mwi(struct pci_dev *pdev)
{
}
#endif

static int lpp_probe(struct pci_dev *pdev, const struct pci_device_id *ent)
{
	struct lpc_dev *dev;
	int rc;
	uint minor;
	unsigned long plx_base;

	dev = kzalloc(sizeof(struct lpc_dev), GFP_KERNEL);
	if (!dev)
		return -ENOMEM;

	/* enable device (incl. PCI PM wakeup), and bus-mastering */
	rc = pci_enable_device(pdev);
	if (rc)
		goto err_out_free;

	rc = pci_set_mwi(pdev);
	if (rc)
		goto err_out_disable;

	rc = pci_request_regions(pdev, DRIVER_MOD_NAME);
	if (rc)
		goto err_out_mwi;

	if (pdev->irq < 2) {
		rc = -EIO;
		dev_err(&pdev->dev, "invalid irq (%d)\n", pdev->irq);
		goto err_out_res;
	}

	plx_base     = pci_resource_start(pdev, 1);
	dev->PlxIntCsrPort = plx_base + P9050_INTCSR;
	dev->cport   = pci_resource_start(pdev, 2);
	dev->rport   = pci_resource_start(pdev, 3);
	dev->wport   = dev->rport;
	dev->wtcport = pci_resource_start(pdev, 5);
	dev->irq     = pdev->irq;

	/* select a "subminor" number (part of a minor number) */
	mutex_lock(&minor_table_mutex);
	for (minor = MAX_LPCS; minor < MAX_DEVICES; ++minor) {
		if (minor_table[minor] == NULL)
			break;
	}
	if (minor >= MAX_DEVICES) {
		dev_info(&pdev->dev, "Too many devices plugged in, can not " \
				"handle this device.");
		rc = -EINVAL;
		goto err_minor_table;
	}

	common_init(dev);

	dev->minor = minor;
	minor_table[minor] = dev;

	pci_set_drvdata(pdev, dev);
	mutex_unlock(&minor_table_mutex);

	device_create(lpcclass, &pdev->dev, MKDEV(major, minor), NULL,
			"lppdrv%u", minor - MAX_LPCS);
	return 0;

err_minor_table:
	mutex_unlock(&minor_table_mutex);
err_out_res:
	pci_release_regions(pdev);
err_out_mwi:
	pci_clear_mwi(pdev);
err_out_disable:
	pci_disable_device(pdev);
err_out_free:
	kfree(dev);
	return rc;
}

static const struct file_operations lpc_fops = {
	.owner =		THIS_MODULE,
	.read =			lpc_read,
	.write =		lpc_write,
	.poll =			lpc_poll,
	.unlocked_ioctl =	lpc_ioctl,
	.open =			lpc_open,
	.release =		lpc_release,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 35)
	.llseek =		noop_llseek,
#endif
};

static void lpp_remove_one(struct pci_dev *pdev)
{
	struct lpc_dev *dev = pci_get_drvdata(pdev);

	mutex_lock(&minor_table_mutex);
	minor_table[dev->minor] = NULL;
	pci_release_regions(pdev);
	pci_clear_mwi(pdev);
	pci_disable_device(pdev);
	pci_set_drvdata(pdev, NULL);
	kfree(dev);
	mutex_unlock(&minor_table_mutex);

	device_destroy(lpcclass, MKDEV(major, dev->minor));
}

static struct pci_driver lpp_driver = {
	.name         = DRIVER_MOD_NAME,
	.id_table     = lpp_pci_tbl,
	.probe        =	lpp_probe,
	.remove       = lpp_remove_one,
};

static int __init module_lpc_init(void)
{
	int i, val, ret;
	char c;

	ret = register_chrdev(major, DRIVER_DEV_NAME, &lpc_fops);
	if (ret < 0)
		return ret;
	major = ret;

	lpcclass = class_create(THIS_MODULE, CLASS_NAME);
	if (IS_ERR(lpcclass)) {
		pr_err("class_create(%s) failed\n", CLASS_NAME);
		ret = PTR_ERR(lpcclass);
		goto out_unregister;
	}

	if (isa) {
		i = 0;
		while (*isa && (i < MAX_LPCS * 4)) {
			c = *isa++;
			val = -1;
			if ((c >= '0') && (c <= '9'))
				val = c - '0';
			if ((c >= 'A') && (c <= 'F'))
				val = c + 10 - 'A';
			if ((c >= 'a') && (c <= 'f'))
				val = c + 10 - 'a';
			if (val >= 0) {
				lpcs[i >> 2] = (lpcs[i >> 2] << 4) | val;
				i++;
			}
		}
	}

	ret = lpc_init();
	if (ret)
		goto out_unregister;

	ret = pci_register_driver(&lpp_driver);
	if (ret)
		goto err_pci;

	return ret;
err_pci:
	cleanup_isa();

out_unregister:
	class_destroy(lpcclass);
	unregister_chrdev(major, DRIVER_DEV_NAME);
	return ret;
}

static void __exit module_lpc_exit(void)
{
	pci_unregister_driver(&lpp_driver);
	cleanup_isa();

	unregister_chrdev(major, DRIVER_DEV_NAME);
	class_destroy(lpcclass);
}
module_init(module_lpc_init);
module_exit(module_lpc_exit);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
