/*
 * Serial LonTalk Adapter (SLTA)
 * Linux kernel serial line discipline implementation
 * Author:	Michael Tross <mtr@nets-solutions.de>,
 *		Volker Schober <vschober@gesytec.de>
 *
 * Derivered from Echelon Solaris SLTA interface lvdslta.c
 * Copyright (c) 1992, 1995 Echelon Corp. All rights reserved.
 * Copyright (c) 2006, 2008 Michael Tross, NETS-Solutions, All Rights Reserved.
 */

/*
 * Configurations
 * LPS-protocol:
 *   - USE_RELIABLE = 1
 *   - slta_alert_ack = 0
 *   - slta_skip_brequest = 1
 *
 * SLTA-protocol:
 *   - USE_RELIABLE = 0
 *   - slta_alert_ack = 1
 *   - slta_skip_brequest = 0
 */

#include <linux/module.h>
#include <linux/slab.h>
#include <linux/tty.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/poll.h>
#include <linux/bitops.h>
#include <linux/sched.h>
#include <linux/version.h>

#include "slta-tty.h"

static int slta_alert_ack = !USE_RELIABLE; /* disables/enables alert ack protocol */
static int slta_skip_brequest = USE_RELIABLE; /* V7 SLTA's may allow this. */
static int slta_buffers = 64; /* sets input/output buffer count */
static int slta_debug = 0;

module_param(slta_alert_ack, int, S_IRUGO | S_IWUSR);
module_param(slta_debug, int, S_IRUGO | S_IWUSR);
module_param( slta_buffers, int, S_IRUGO);

MODULE_PARM_DESC(slta_alert_ack, "Enables Alert-Ack-Protocol");
MODULE_PARM_DESC(slta_buffers, "I/O buffer count");
MODULE_PARM_DESC(slta_debug, "Debugging Level");

MODULE_ALIAS_LDISC(N_SLTA);

static const u8 c_ack_rsp = RM_ACK;
static const u8 c_nack_rsp = RM_NACK;

/* pending receive datagrams */
static int slta_rx_ready;

/* transmit bottom-half handler */
static void slta_transmit_guard(unsigned long arg);
#ifdef DECLARE_TASKLET_OLD
// !!! (Lue) since 5.10 just two parametes
static DECLARE_TASKLET_OLD(sltatx_tasklet, slta_transmit_guard);
#else
static DECLARE_TASKLET(sltatx_tasklet, slta_transmit_guard, 0);
#endif

/*
 * This is the timer signal processing function. This function is
 * set up to provide a once-a-second service to decrement certain local
 * timers.
 */
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
static void slta_timer_service(unsigned long p)
{
	struct slta_dev *dev = (struct slta_dev *) p;
#else
void slta_timer_service(struct timer_list *t)
{
    struct slta_dev *dev = from_timer(dev, t, slta_timer);
#endif	

	if (dev->ack_wait_timer)
		dev->ack_wait_timer--;

	if (dev->rxstate_tmo)
	{
		if (--dev->rxstate_tmo == 0)
		{
			/* abort the input processing */
			SLTA_DEBUG(3, "%s(): rx_state=%d -> idel\n", __func__, dev->rx_state);
			dev->rx_state = MSRX_IDLE;
		}
	}

	if (dev->txwait1_tmo)
	{
		if (--dev->txwait1_tmo == 0)
		{
			/* continue header output processing -
			 probably slta not in ALERT_ACK mode ?! */
			if (slta_alert_ack)
			{
				SLTA_DEBUG(3, "%s(): tx_state=%d -> idle\n", __func__, dev->tx_state);
				dev->tx_state = MSTX_IDLE;
				dev->bpt_count = 0;
			}
#if USE_RELIABLE
			if (dev->tx_state != MSTX_IDLE || dev->sxo_tail->busy)
#endif
			tasklet_schedule(&sltatx_tasklet);
		}
	}

	dev->slta_timer.expires = jiffies + msecs_to_jiffies(TIMER_INTERVALL);
	add_timer(&dev->slta_timer);
}

#if 0
/*
 * Put a niFLUSH_CANCEL command on the output list. This is
 * done at open time and whenever uplink niRESETs are encountered.
 */
static void slta_do_flush_cancel(struct slta_dev *dev)
{
	spin_lock(&dev->sxo_lock);

	dev->sxo_head->llh.ni_cmd = niFLUSH_CANCEL;
	dev->sxo_head->llh.length = 1;
	dev->sxo_head->busy = true;
	dev->sxo_head = dev->sxo_head->next;

	spin_unlock(&dev->sxo_lock);
}

/*
 * Put a niRESET command on the output list.
 */
static void slta_do_reset(struct slta_dev *dev)
{
	spin_lock(&dev->sxo_lock);

	dev->sxo_head->llh.ni_cmd = niRESET;
	dev->sxo_head->llh.length = 1;
	dev->sxo_head->busy = true;
	dev->sxo_head = dev->sxo_head->next;

	spin_unlock(&dev->sxo_lock);
}
#endif

/**
 *	sltatty_set__room	-	receive space
 *	@tty: terminal
 *
 *	Called by the driver to find out how much data it is
 *	permitted to feed to the line discipline without any being lost
 *	and thus to manage flow control. Not serialized. Answers for the
 *	"instant".
 */
static void sltatty_set_room(struct tty_struct *tty)
{
	int left = (slta_buffers - slta_rx_ready) * sizeof(Ibuffer) - 1;
	tty->receive_room = left;

	SLTA_DEBUG(3, "%s(): receive room = %d\n", __func__, left);
}

/*
 * Enqueue the received message onto the input buffer list.
 */
static void slta_msg_enqueue(struct slta_dev *dev, int if_error)
{
	struct sltatty_cb *priv = dev->priv;
	struct tty_struct *tty;
	unsigned long flags;

	if (if_error)
		SLTA_DEBUG(3, "%s(): if_error\n", __func__);

	SLTA_ASSERT(priv != NULL, return;);
	SLTA_ASSERT(priv->magic == SLTATTY_MAGIC, return;);

	tty = priv->tty;

	spin_lock_irqsave(&dev->sxi_lock, flags);
	if (!dev->sxi_head->busy)
	{
		dev->sxi_head->busy = true;
		dev->sxi_head->error = if_error;
		dev->sxi_head = dev->sxi_head->next;

		/* avoid rx-buffer overflow condition: drop the oldest packet */
		if (dev->sxi_head->busy)
		{
			SLTA_DEBUG(2, "%s: rx overflow\n", __func__);
			dev->sxi_head->busy = false;
			dev->sxi_head->error = 0;
			dev->sxi_tail = dev->sxi_tail->next;
		}
		else
		{
			slta_rx_ready++;
			sltatty_set_room(tty);
		}
	}
	spin_unlock_irqrestore(&dev->sxi_lock, flags);

	wake_up_interruptible(&dev->rd_wait);
	kill_fasync(&tty->fasync, SIGIO, POLL_IN);
}

/*
 * Dequeue the current write message, advance the write buffer head
 */

static void slta_msg_dequeue(struct slta_dev *dev)
{
	struct sltatty_cb *priv = dev->priv;
	struct tty_struct *tty;
	unsigned long flags;

	SLTA_ASSERT(priv != NULL, return;);
	SLTA_ASSERT(priv->magic == SLTATTY_MAGIC, return;);

	tty = priv->tty;

	spin_lock_irqsave(&dev->sxo_lock, flags);

	dev->sxo_tail->busy = false;
	dev->sxo_tail = dev->sxo_tail->next;

	spin_unlock_irqrestore(&dev->sxo_lock, flags);

	wake_up_interruptible(&dev->wr_wait);
	kill_fasync(&tty->fasync, SIGIO, POLL_OUT);
}

/*
 * Initialize the forward list links for the input buffers
 */

static void slta_msg_reset(struct slta_dev *dev)
{
	struct sltatty_cb *priv = dev->priv;
	int i;
	Ibuffer *np;

	/* Initialize the input buffer list. */
	spin_lock_irq(&dev->sxi_lock);

	i = slta_buffers;
	dev->sxi_tail = dev->sxi_head = np = (Ibuffer *) dev->heap_ptr;
	while (i--)
	{
		SLTA_DEBUG(5, "  in %2u -> @=%p\n", i, dev->sxi_head);
		dev->sxi_head->busy = false;
		np = dev->sxi_head + 1;

		/* Last one links back to first. */
		dev->sxi_head->next = (i) ? np : dev->sxi_tail;
		dev->sxi_head = dev->sxi_head->next;
	}
	SLTA_DEBUG(4, "  in: sxi_head=%p, sxi_tail=%p\n", dev->sxi_head,
			dev->sxi_tail);

	spin_unlock_irq(&dev->sxi_lock);

	/* Next build the output buffer list. */
	spin_lock_irq(&dev->sxo_lock);

	i = slta_buffers;
	dev->sxo_tail = dev->sxo_head = np;
	while (i--)
	{
		SLTA_DEBUG(5, "  out %2u -> @=%p\n", i, dev->sxo_head);
		dev->sxo_head->busy = false;

		/* Last one links back to first. */
		dev->sxo_head->next = (i) ? dev->sxo_head + 1 : dev->sxo_tail;
		dev->sxo_head = dev->sxo_head->next;
	}
	SLTA_DEBUG(4, "  out: sxo_head=%p, sxo_tail=%p\n", dev->sxo_head,
			dev->sxo_tail);

	slta_rx_ready = 0;
	sltatty_set_room(priv->tty);

	spin_unlock_irq(&dev->sxo_lock);

	/* write ressources are free now - wakeup anyone waiting */
	wake_up_interruptible(&dev->wr_wait);
	kill_fasync(&priv->tty->fasync, SIGIO, POLL_OUT);
}

/* device configuration callbacks always invoked with slta-thread context */

#if 0
/* find out, how many chars we have in buffers below us
 * this is allowed to lie, i.e. return less chars than we
 * actually have. The returned value is used to determine
 * how long the sltathread should wait before doing the
 * real blocking wait_until_sent()
 */
static int sltatty_chars_in_buffer(struct slta_dev *dev)
{
	struct sltatty_cb *priv = dev->priv;
	int num;

	SLTA_ASSERT(priv != NULL, return -1;);
	SLTA_ASSERT(priv->magic == SLTATTY_MAGIC, return -1;);

	num = tty_chars_in_buffer(priv->tty);

	SLTA_DEBUG(3, "%s(): chars in buffer = %d\n", __func__, num);
	return num;
}

/* Wait (sleep) until underlaying hardware finished transmission
 * i.e. hardware buffers are drained
 * this must block and not return before all characters are really sent
 *
 * If the tty sits on top of a 16550A-like uart, there are typically
 * up to 16 bytes in the fifo - f.e. 9600 bps 8N1 needs 16.7 msec
 *
 * With usbserial the uart-fifo is basically replaced by the converter's
 * outgoing endpoint buffer, which can usually hold 64 bytes (at least).
 * With pl2303 it appears we are safe with 60msec here.
 *
 * I really wish all serial drivers would provide
 * correct implementation of wait_until_sent()
 */
static void sltatty_wait_until_sent(struct slta_dev *dev)
{
	struct sltatty_cb *priv = dev->priv;
	struct tty_struct *tty;

	SLTA_ASSERT(priv != NULL, return;);
	SLTA_ASSERT(priv->magic == SLTATTY_MAGIC, return;);

	tty = priv->tty;

	if (tty->ops->wait_until_sent)
	tty->ops->wait_until_sent(tty, msecs_to_jiffies(100));
}
#endif

/*
 *  Function sltatty_change_speed (dev, speed)
 *
 *    Change the speed of the serial port.
 *
 * This may sleep in set_termios (usbserial driver f.e.) and must
 * not be called from interrupt/timer/tasklet therefore.
 * All such invocations are deferred to kSltad now so we can sleep there.
 */

static int sltatty_change_speed(struct slta_dev *dev, unsigned speed)
{
	struct sltatty_cb *priv = dev->priv;
	struct tty_struct *tty;
	struct tty_driver *drv;
	struct ktermios old_termios;
	int cflag;

	SLTA_ASSERT(priv != NULL, return -1;);
	SLTA_ASSERT(priv->magic == SLTATTY_MAGIC, return -1;);

	tty = priv->tty;

	drv = NULL;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,12, 0)
	down_write(&tty->termios_rwsem);
#else
	mutex_lock(&tty->termios_mutex);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0)
	old_termios = tty->termios;
	cflag = tty->termios.c_cflag;
#else
	old_termios = *(tty->termios);
	cflag = tty->termios->c_cflag;
#endif

	tty_encode_baud_rate(tty, speed, speed);
	if (tty->ops->set_termios)
		tty->ops->set_termios(tty, &old_termios);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,12, 0)
	up_write(&tty->termios_rwsem);
#else
	mutex_unlock(&tty->termios_mutex);
#endif

	dev->speed = speed;

	return 0;
}

#if 0
static int sltatty_set_dtr_rts(struct slta_dev *dev, int dtr, int rts)
{
	struct sltatty_cb *priv = dev->priv;
	int set = 0;
	int clear = 0;

	SLTA_ASSERT(priv != NULL, return -1;);
	SLTA_ASSERT(priv->magic == SLTATTY_MAGIC, return -1;);

	if (rts)
	set |= TIOCM_RTS;
	else
	clear |= TIOCM_RTS;
	if (dtr)
	set |= TIOCM_DTR;
	else
	clear |= TIOCM_DTR;

	SLTA_DEBUG(2, "%s(): set dtr %d, rts %d - IGNORED\n", __func__, dtr, rts);
	/*
	 * We can't use ioctl() because it expects a non-null file structure,
	 * and we don't have that here.
	 * This function is not yet defined for all tty driver, so
	 * let's be careful... Jean II
	 */
	return 0;
}
#endif

/* called from slta_dev when there is more data to send
 * context is either netdev->hard_xmit or some transmit-completion bh
 * i.e. we are under spinlock here and must not sleep.
 */

static int sltatty_do_write(struct slta_dev *dev, const unsigned char *ptr, size_t len)
{
	struct sltatty_cb *priv = dev->priv;
	struct tty_struct *tty;
	struct tty_driver *drv;
	int writelen, room;

	SLTA_ASSERT(priv != NULL, return -1;);
	SLTA_ASSERT(priv->magic == SLTATTY_MAGIC, return -1;);

	tty = priv->tty;
	drv = tty->driver;

	tty->flags |= (1 << TTY_DO_WRITE_WAKEUP);

	writelen = room = tty_write_room(tty);

	if (writelen > len)
		writelen = len;

	SLTA_DEBUG(12, "%s(): %p, %zu bytes (%u free)\n", __func__, ptr, len, room);

	if (slta_debug > 3)
		DumpBuffer(">>", (void *) ptr, writelen);

	if (writelen == 1)
	{
		tty_put_char(tty, *ptr);
		return 1;
	}

	if (!tty->ops->write)
	{
		SLTA_WARNING("%s(): tty driver has no write handler!\n", __func__);
		return 0;
	}
	return tty->ops->write(tty, ptr, writelen);
}

/*
 * Write to the tty device, adjust the pointer and count.
 * Return number of bytes written.
 */
static int slta_dl_write(struct slta_dev *dev, int length, bool *tx_contin)
{
	int retval, i;

	SLTA_DEBUG(5, "%s() rx_state=%d tx_state=%d mip_state=%d "
			"bpt_count=%d\n", __func__, dev->rx_state,
			dev->tx_state, dev->mip_state, dev->bpt_count);

	/* Always add in the checksum
	 ** (although not required when sending header). */
	for (i = 0; i < length; i++)
		dev->mstx_chksum += dev->bpt[i];

	retval = sltatty_do_write(dev, dev->bpt, length);

	if (retval > 0)
	{
		dev->bpt += retval;
		dev->bpt_count -= retval;
		SLTA_DEBUG(13, "%s: wrote %u out of %u bytes\n",
				__func__, retval, length);
	}

	/* Clear continuation flag if output device is full. */
	if (retval != length)
		*tx_contin = false;

	return retval;
}

/*
 * Function slta_do_flush (struct slta_dev)
 *
 *    Flush output buffer.
 *
 */
static void sltatty_do_flush(struct slta_dev *dev)
{
	struct sltatty_cb *priv = dev->priv;
	struct tty_struct *tty = priv->tty;

	if (tty->ops->flush_chars)
		tty->ops->flush_chars(tty);
}

/*
 * Function slta_write_flush (tty)
 *
 *    Called by the internal state machine when the transmit buffer
 *    should be sent immediately.
 *
 */
static void slta_write_flush(struct tty_struct *tty)
{
	if (tty->ops->flush_chars)
		tty->ops->flush_chars(tty);
}

/*
 * This procedure services the tty transmit buffer.
 * The network interface link layer header is copied from the buffer
 * area to a local header structure <dl_head>. From there is is sent
 * to the tty device. The data portion is taken directly from the buffer area.
 * The checksum is sent last. When the downlink traffic requires a buffer
 * request then one is built and sent downlink here.
 * This procedure may be called repeatedly until the return value becomes
 * false, indicating that the output device is full, or the transmit state
 * is waiting for uplink (receive) data.
 */
static bool slta_tx_process(struct slta_dev *dev, u8 AckNack)
{
	struct sltatty_cb *priv;
	struct tty_struct *tty;
	struct tty_driver *drv;
	u8 l_ni_type;
	bool tx_contin;
	static u8 l_mstx_chksum;
	static int brq_count;

	priv = dev->priv;
	tty = priv->tty;
	drv = tty->driver;

#if !USE_RELIABLE
	/* If the receiver is busy then back off for now. */
	if (dev->rx_state != MSRX_IDLE)
	{
		SLTA_DEBUG(3, "%s: rx state 0x%X not idle\n", __func__, dev->rx_state);
		return false;
	}
#endif

	/* may get set to false by dl_write() */
	tx_contin = true;

	switch (dev->tx_state)
	{

	case MSTX_IDLE:

		if (dev->rx_state != MSRX_IDLE)
		{
			SLTA_DEBUG(3, "%s: rx state %d, not idle\n", __func__,
					dev->rx_state);
			return false;
		}

		if (dev->txwait1_tmo)
		{
			SLTA_DEBUG(3, "%s: tx timeout active, txwait1_tmo %d\n",
					__func__, dev->txwait1_tmo);
			break;
		}

		/* Check for pending downlink traffic. Skip if our input
		 ** buffers are clogged, or if already set up! */
		if (dev->sxo_tail->busy && !dev->sxi_head->busy && dev->bpt_count == 0)
		{
			/* Set up for start of downlink transfer. */
			dev->dl_head.alert = NI_ALERT;
			dev->dl_head.ni_cmd = dev->sxo_tail->llh.ni_cmd;
			dev->dl_head.length = dev->sxo_tail->llh.length;
#if USE_RELIABLE
			dev->dl_head.seqn = dev->transmit_sequence_n;
#endif
			SLTA_DEBUG(3, "%s: new packet(%d) @%p, cmd 0x%x, len %u\n",
					__func__, dev->transmit_sequence_n, dev->sxo_tail,
					dev->dl_head.ni_cmd, dev->dl_head.length);

			/*
			 * Is this a downlink message requiring an output
			 * buffer?
			 */
			l_ni_type = dev->dl_head.ni_cmd & niTYPE;
			if (l_ni_type == niCOMM || l_ni_type == niNETMGMT)
			{
				if (slta_skip_brequest)
					goto DoAck_Okay;
				switch (dev->mip_state)
				{
				case MIP_ACK_WAIT:
					/*
					 * Still waiting for a buffer request
					 * ack. Check for a timeout.
					 */
					if (dev->ack_wait_timer)
						return false;
					/* FALLTHRU */
				case MIP_ACK_NACK:
					/*
					 * If the d.l. buffer requests
					 * eventually fail, indicate a failure.
					 */
					if (--brq_count == 0)
					{
						dev->g_tx_error = true;
						dev->mip_state = MIP_IDLE;
						return false;
					}
					/* FALLTHRU */
				case MIP_IDLE:
					/* Start a downlink buffer request. */
					if (brq_count == 0)
						brq_count = BRQ_RETRIES;
					dev->ack_wait_timer = ACK_WAIT_PD;
					dev->mip_state = MIP_ACK_WAIT;
					dev->dl_head.length = 1;
					dev->dbg.brqs_sent++;
					break;
				case MIP_ACK_OKAY:
					DoAck_Okay:
					/* Got buffer req. ACK, send message. */
					brq_count = 0;
					dev->dl_head.length = dev->sxo_tail->llh.length;
					dev->mip_state = MIP_IDLE;
					dev->dbg.mbfs_sent++;
					break;
				}
			}
			dev->dl_head.nlength = ~dev->dl_head.length;
			dev->bpt = (u8 *) &dev->dl_head;
			dev->bpt_count = sizeof(dev->dl_head);

			SLTA_DEBUG(3, "%s: start xmit from %p, len %u\n",
					__func__, dev->bpt, dev->bpt_count);

		}

		/* Now transmit as much of the header as we can.
		 ** For the alert_ack_prtcl case only send the niALERT,
		 ** otherwise the whole header can be sent. If nothing
		 ** can be sent at this point then stay in the MSTX_IDLE
		 ** state. */
		if (dev->bpt_count)
		{
			if (dev->bpt_count == sizeof(dev->dl_head) &&
					slta_alert_ack)
			{
				if (slta_dl_write(dev, 1, &tx_contin) > 0)
					dev->tx_state = MSTX_WAIT1;
				dev->txwait1_tmo = TXWAIT1_TMO;
			}
			else
			{
				if (slta_dl_write(dev, dev->bpt_count, &tx_contin) > 0)
					dev->tx_state = MSTX_TXHEAD;
			}

			/* Immediately sent the Alert/Header bytes(s) */
			slta_write_flush(tty);
		}
		break;

	case MSTX_WAIT1:
		SLTA_DEBUG(3, "%s: state mstx_wait1\n", __func__);
		break;

	case MSTX_HOLD:
		/* Do not a thing. */
		SLTA_DEBUG(3, "%s: state mstx_txhold\n", __func__);
		break;

	case MSTX_TXHEAD:
		SLTA_DEBUG(3, "%s: state mstx_txhead, from %p, len %u\n",
				__func__, dev->bpt, dev->bpt_count);

		/* Stop alert/ack timeout handler */
		dev->txwait1_tmo = 0;

		if (dev->bpt_count)
			slta_dl_write(dev, dev->bpt_count, &tx_contin);
		if (dev->bpt_count == 0)
		{
			/*
			 * Header is sent, reset mstx_chksum and check if there
			 * is a data portion.
			 */
			dev->mstx_chksum = dev->dl_head.ni_cmd;
			SLTA_DEBUG(3, "%s: dl_head.length=%u\n",
					__func__, dev->dl_head.length);
			if (dev->dl_head.length > 1)
			{
				dev->bpt = dev->sxo_tail->llm.ni_data;

				dev->bpt_count = dev->dl_head.length - 1;

#if !USE_RELIABLE
				/*
				 * For the buffered_prtcl case flush the output,
				 * assuring that (most of) the header has been
				 * sent, process the receiver and check its
				 * state, and if not idle, hold off TXDATA state
				 * until it is. We only do this when sending
				 * downlink messages rather than command only
				 * transfers.
				 */
				if (!slta_alert_ack)
				{
					slta_write_flush(tty);

					dev->tx_state = (dev->rx_state == MSRX_IDLE) ?
															MSTX_TXDATA :
															MSTX_HOLD;
				}
				else
				#endif
				{
					dev->tx_state = MSTX_TXDATA;
					goto DoTxData;
				}
			}
			else
			{
				dev->tx_state = MSTX_TXCS;
				goto DoTxCS;
			}
		}
		SLTA_DEBUG(3, "%s: state mstx_txhead, next tx_state=%u\n",
				__func__, dev->tx_state);
		break;

	case MSTX_TXDATA:
		DoTxData:
		SLTA_DEBUG(3, "%s: state mstx_txdata, from %p, len %u\n",
				__func__, dev->bpt, dev->bpt_count);

		slta_dl_write(dev, dev->bpt_count, &tx_contin);
		if (dev->bpt_count == 0)
		{
			dev->tx_state = MSTX_TXCS;
			goto DoTxCS;
		}
		break;

	case MSTX_TXCS:
		DoTxCS:
		SLTA_DEBUG(3, "%s: state mstx_txcs\n", __func__);

		/* No more data, send checksum. */
		l_mstx_chksum = -dev->mstx_chksum;
		dev->bpt = &l_mstx_chksum;
		dev->bpt_count = sizeof(dev->mstx_chksum);
		slta_dl_write(dev, dev->bpt_count, &tx_contin);
		slta_write_flush(tty);

#if USE_RELIABLE
		/* If CS is sent then change state */
		dev->tx_state = MSTX_TXDONE;
		dev->txwait1_tmo = TXWAIT2_TMO;
#else
		/* If CS is sent then change state */
		dev->tx_state = MSTX_IDLE;
		dev->txwait1_tmo = 0;

		/* If MIP_IDLE then free up the output buffer */
		if (dev->mip_state == MIP_IDLE)
		{
			SLTA_DEBUG(3, "%s: transmit complete(%02X) @%p, next @%p\n",
					__func__, dev->transmit_sequence_n, dev->sxo_tail, dev->sxo_tail->next);
			slta_msg_dequeue(dev);

		}
		else
		{
			SLTA_DEBUG(3, "%s: transmit complete(%02X) @%p, MIP != IDLE\n",
					__func__, dev->transmit_sequence_n, dev->sxo_tail);
		}
#endif
		SLTA_DEBUG(3, "%s: state mstx_txcs, next tx_state=%u\n",
				__func__, dev->tx_state);
		break;

#if USE_RELIABLE
	case MSTX_TXDONE:
		if (AckNack == 0 && dev->txwait1_tmo)
		break;

		/* If ACK/NACK is received */
		dev->tx_state = MSTX_IDLE;
		dev->txwait1_tmo = 0;

		if (AckNack == RM_ACK)
		{
			/* If MIP_IDLE then free up the output buffer */
			if (dev->mip_state == MIP_IDLE)
			{
				SLTA_DEBUG(3, "%s: transmit complete(%02x) "
						"@%p, next @%p\n", __func__,
						dev->transmit_sequence_n,
						dev->sxo_tail,
						dev->sxo_tail->next);
				ClearWrBuf:
				slta_msg_dequeue(dev);

				dev->transmit_sequence_n++;
				if (!dev->transmit_sequence_n)
				dev->transmit_sequence_n = 1;

				dev->SendRetries = MAX_SEND_RETRIES;
				dev->SendNackRetries = MAX_SEND_NACK_RETRIES;
			}
			else
			{
				SLTA_DEBUG(3, "%s: transmit complete(%02x) "
						"@%p, MIP != IDLE\n", __func__,
						dev->transmit_sequence_n,
						dev->sxo_tail);
			}

		}
		else if (AckNack == RM_NACK)
		{
			SLTA_DEBUG(3, "%s: transmit complete(%02x) @%p, "
					"SendNackRetries=%d AckNack == RM_NACK\n",
					__func__, dev->transmit_sequence_n,
					dev->sxo_tail, dev->SendNackRetries);
			if (--dev->SendNackRetries)
			{
				dev->txwait1_tmo = SEND_NACK_REPEAT_DELAY;
				break;
			}
			goto ClearWrBuf;

		}
		else if (AckNack == 0)
		{
			/* Timeout */
			SLTA_DEBUG(3, "%s: transmit complete(%02x) @%p, "
					"SendRetries=%d Timeout\n", __func__,
					dev->transmit_sequence_n,
					dev->sxo_tail, dev->SendRetries);
			if (--dev->SendRetries)
			{
				dev->txwait1_tmo = SEND_REPEAT_DELAY;
				break;
			}
			goto ClearWrBuf;
		}

		if (dev->sxo_tail->busy)
		tasklet_schedule(&sltatx_tasklet);

		break;
#endif
	}

	/* Check for the need to break the continuation of this procedure. */
	if (dev->tx_state == MSTX_WAIT1 || dev->tx_state == MSTX_HOLD ||
			/* true IDLE state? */
			(dev->tx_state == MSTX_IDLE && dev->bpt_count == 0))
		return false;

	return tx_contin;
}

static bool handle_rxstate_rxidle(struct slta_dev *dev)
{
	u8 ni_byte;

	SLTA_DEBUG(3, "%s: state msrx_idle\n", __func__);

	/* START of alert_ack_prtcl code. */
	dev->rx_try_flag = false;

	/* If in the MSTX_WAIT1 state then we are using the alert_ack_prtcl,
	 ** and should receive an NI_ALERT_ACK. If so, change the tx state
	 ** and exit. */
	if (dev->tx_state == MSTX_WAIT1)
	{
		if (dev->sxi_head->llh.alert == NI_ALERT_ACK)
		{
			SLTA_DEBUG(3, "%s:  -> rcvd alert ack, commence xmit header\n", __func__);
			dev->tx_state = MSTX_TXHEAD;
			dev->dbg.alacks_rcvd++;
			return true;

		}
		else if (dev->sxi_head->llh.alert == NI_ALERT)
		{
			SLTA_DEBUG(3, "%s:  -> rcvd alert, collision, abort xmit\n", __func__);
			/*
			 * May have received an NI_ALERT from the SLTA. If
			 * this is the case then a 'collision' is occuring.
			 * Blow off the downlink transfer. The NI_ALERT that we
			 * sent will result in the SLTA sending another
			 * NI_ALERT.
			 */
			dev->tx_state = MSTX_IDLE;
			dev->bpt_count = 0;
			/* Go to a special state to hold off re-tx'ing */
			dev->rx_state = MSRX_CLASH;
			if (dev->mip_state == MIP_ACK_WAIT)
				dev->mip_state = MIP_IDLE;
			dev->dbg.clashes++;
			return true;
		}
	}
	/* END of alert_ack_prtcl code. */

	/* If MSRX_IDLE, and there is input, then start a timer and fall into
	 * MSRX_RXHEAD. For the alert_ack_prtcl case, flush any other NI_ALERTs
	 * that may have queued up, send a downlink NI_ALERT_ACK, and change
	 * state.
	 */
	if (slta_alert_ack)
	{
		if (!(dev->sxi_head->llh.alert == NI_ALERT))
		{
			/* unknown uplink front end. */
			/* stay in MSRX_IDLE. */
			return true;
		}
		/*
		 * empty the receiver SILO. there could be several NI_ALERTs
		 * in there if this process has been ignored for a while.
		 */
		dev->rxb_drop = 1;

		/* Transmit the NI_ALERT_ACK */
		ni_byte = NI_ALERT_ACK;
		sltatty_do_write(dev, &ni_byte, 1);
		sltatty_do_flush(dev);

		/* Assure that the Llhed pointer is correct. */
		dev->rxb_indx = 1;
		dev->rxb_ptr = ((u8 *) &dev->sxi_head->llh) + 1;
	}
#if USE_RELIABLE
	if (dev->tx_state != MSTX_IDLE)
	{
		SLTA_DEBUG(3, "%s: tx state not idle, rcvd %02x\n",
				__func__, dev->sxi_head->llh.alert);
		if (dev->sxi_head->llh.alert == RM_ACK ||
				dev->sxi_head->llh.alert == RM_NACK)
		{
			slta_tx_process(dev, dev->sxi_head->llh.alert);
		}
	}
	if (dev->sxi_head->llh.alert != NI_ALERT)
	return true; /* stay in MSRX_IDLE. */
#endif
	dev->rxstate_tmo = RXWAIT1_TMO;
	dev->rx_state = MSRX_RXHEAD;
	return false;
}

static void handle_rxstate_rxhead(struct slta_dev *dev)
{
	SLTA_DEBUG(3, "%s: state msrx_head\n", __func__);

	/* Check for completion of input of the Llhed. */
	if (!(dev->rxb_indx == sizeof(Llhed)))
		return;

	/* Validate the Llhed */
	if (dev->sxi_head->llh.alert == NI_ALERT &&
			dev->sxi_head->llh.length ==
					(dev->sxi_head->llh.nlength ^ 0xFF))
	{
		dev->rx_state = MSRX_RXDATA;
		return;
	}
	/*
	 * Bad Llhed content. In the alert_ack_prtcl case this could be due to
	 * one extra NI_ALERT in the input stream. This could happen when this
	 * process gets switched out just after the above "empty" read code,
	 * and the SLTA sends another NI_ALERT. For this case try shifting out
	 * one byte in the header, and let the next go-round finish the Llhed
	 * input.
	 */
	if (slta_alert_ack && !dev->rx_try_flag)
	{
		dev->sxi_head->llh.length = dev->sxi_head->llh.nlength;
		dev->sxi_head->llh.nlength = dev->sxi_head->llh.ni_cmd;
		dev->rxb_indx--;
		dev->rxb_ptr--;
		dev->rx_try_flag = true;
		return;
	}
	/* Else, log an error. */
	slta_msg_enqueue(dev, true);
	/* empty the receiver SILO */
	/* read(tt_handle, sxi_head->llm.ni_data, MAX_NI_DATA+1); */
	dev->rxb_drop = 1;
	dev->rx_state = MSRX_IDLE;
}

static void handle_rxstate_rxdata(struct slta_dev *dev)
{
	int csum, i;
	u8 oldseqn;

	SLTA_DEBUG(3, "%s: state msrx_data, idx %u, expected %u + header %zu\n",
			__func__, dev->rxb_indx, dev->sxi_head->llh.length, sizeof(Llhed));

	/* Check for completion of input - long datagrams are likely split into pieces */
	if (dev->rxb_indx < (sizeof(Llhed) + dev->sxi_head->llh.length))
	{
		SLTA_DEBUG(3, "%s: incomplete, idx %u, expected %u + "
				"header %zu\n", __func__, dev->rxb_indx, dev->sxi_head->llh.length, sizeof(Llhed));
		return;
	}
	/* Check and deliver ... */
	/* Validate the checksum */
	csum = dev->sxi_head->llh.ni_cmd;
	for (i = 0; i < dev->sxi_head->llh.length; i++)
		csum += dev->sxi_head->llm.ni_data[i];
	if ((csum & 0xFF) != 0)
	{
		SLTA_DEBUG(2, "%s: BAD checksum!\n", __func__);
		slta_msg_enqueue(dev, true);
#if USE_RELIABLE
		sltatty_do_write(dev, &c_nack_rsp, 1);
		sltatty_do_flush(dev);
#endif
		goto out;
	}
#if USE_RELIABLE
	SLTA_DEBUG(2, "%s: datagram complete (%02X->%02X), length %d, "
			"uplink cmd=%02X, mip_state=%u\n", __func__,
			dev->receive_sequence_n,
			dev->sxi_head->llh.seqn,
			dev->sxi_head->llh.length,
			dev->sxi_head->llh.ni_cmd,
			dev->mip_state);

	sltatty_do_write(dev, &c_ack_rsp, 1);
	sltatty_do_flush(dev);

	oldseqn = dev->receive_sequence_n;
	dev->receive_sequence_n = dev->sxi_head->llh.seqn;

	if (dev->sxi_head->llh.seqn && dev->sxi_head->llh.seqn == oldseqn)
	{
		SLTA_ERROR("%s: drop rx packet, invalid sequence number!\n",
				__func__);
		slta_msg_enqueue(dev, true);

		/*
		 * TODO: shouldn't we send a NACK instead?
		 */
		goto out;
	}
#endif
	/* Check the content of the uplink message. */
	switch (dev->sxi_head->llh.ni_cmd)
	{
	case niACK:
		dev->dbg.bracks_rcvd++;
		dev->ack_wait_timer = 0;
		if (dev->mip_state == MIP_ACK_WAIT)
			dev->mip_state = MIP_ACK_OKAY;
		else
			slta_msg_enqueue(dev, false);
		break;
	case niNACK:
		dev->dbg.brnacks_rcvd++;
		dev->ack_wait_timer = 0;
		if (dev->mip_state == MIP_ACK_WAIT)
			dev->mip_state = MIP_ACK_NACK;
		else
			slta_msg_enqueue(dev, false);
		break;
	default:
		slta_msg_enqueue(dev, false);
	}
	out:
	dev->rx_state = MSRX_IDLE;
}

/*
 * This procedure services the serial receive buffer: Firstly it offloads the
 * system's serial input 'silo' to our local buffers and processes buffer
 * ACK/NACKs. This function needs to be executed periodically by the forground
 * program in order to prevent the input 'silo' from overflowing. Alternatively,
 * this function could be executed as a result of an uplink serial interrupt.
 * Since this is a serial device we may not always read in the number of bytes
 * expected at this time, so this is a repeated process that must be
 * executed periodically.
 * Returns true if continued rx processing may be required.
 */
static bool slta_rx_process(struct slta_dev *dev)
{
	struct sltatty_cb *priv;
	struct tty_struct *tty;
	struct tty_driver *drv;
	bool exit;

	priv = dev->priv;
	tty = priv->tty;
	drv = tty->driver;

	switch (dev->rx_state)
	{
	case MSRX_CLASH:
		SLTA_DEBUG(3, "%s: state msrx_clash\n", __func__);

		dev->rx_state = MSRX_IDLE;
		/* FALLTHRU */

	case MSRX_IDLE:
		exit = handle_rxstate_rxidle(dev);
		if (exit)
			return true;
		/* FALLTHRU */

	case MSRX_RXHEAD:
		handle_rxstate_rxhead(dev);
		break;

	case MSRX_RXDATA:
		handle_rxstate_rxdata(dev);
		break;
	}
	return true;
}

/*
 *  Function sltatty_receive_buf( tty, cp, count)
 *
 *    Handle the 'receiver data ready' interrupt.  This function is called
 *    by the 'tty_io' module in the kernel when a block of slta data has
 *    been received, which can now be decapsulated and delivered for
 *    further processing
 *
 * calling context depends on underlying driver and tty->low_latency!
 * for example (low_latency: 1 / 0):
 * serial.c:	uart-interrupt / softint
 * usbserial:	urb-complete-interrupt / softint
 */
static void sltatty_receive_buf(struct tty_struct *tty, const unsigned char *cp, char *fp, int count)
{
	struct slta_dev *dev;
	struct sltatty_cb *priv = tty->disc_data;
	int r_length;

	SLTA_ASSERT(priv != NULL, return;);
	SLTA_ASSERT(priv->magic == SLTATTY_MAGIC, return;);

	if (unlikely(count == 0)) /* yes, this happens */
		return;

	dev = priv->dev;
	if (!dev)
	{
		SLTA_WARNING("%s(), not ready yet!\n", __func__);
		return;
	}

	r_length = count;
	while (count > 0)
	{
		/*
		 * Calculate the size of the read request. If we are in the
		 * process of pulling in the link-layer header, then limit the
		 * read to the size of that header. If we are pulling in the
		 * data portion then limit the read size to the expected length
		 * of the message. This is all just so we don't read into
		 * another uplink message if one is in the input silo.
		 */
		if (dev->sxi_head->busy)
		{
			SLTA_WARNING("%s(), receive buffers full!\n", __func__);
			break;
		}

		switch (dev->rx_state)
		{
		case MSRX_IDLE:
			case MSRX_CLASH:
			dev->rxb_indx = 0;
			dev->rxb_ptr = (u8 *) &dev->sxi_head->llh;
			r_length = 1;
			break;
		case MSRX_RXHEAD:
			r_length = sizeof(Llhed) - dev->rxb_indx;
			break;
		case MSRX_RXDATA:
			r_length = dev->sxi_head->llh.length - (dev->rxb_indx - sizeof(Llhed));
			break;
		}

		/* copy from tty flip buffer */
		if (r_length > count)
			r_length = count;

		memcpy(dev->rxb_ptr, cp, r_length);

		if (slta_debug > 3)
			DumpBuffer("<<", (void *) cp, r_length);

		cp += r_length;
		count -= r_length;

		dev->rxb_indx += r_length;
		dev->rxb_ptr += r_length;

		SLTA_DEBUG(3, "%s(), rx'ed %d bytes!\n", __func__, r_length);

		/* run through the receive state machine */
		slta_rx_process(dev);

		if (dev->rxb_drop)
		{
			SLTA_WARNING("%s(), rxb drop!\n", __func__);
			dev->rxb_drop = 0;
			break;
		}

	}

	if (dev->rx_state == MSRX_IDLE)
	{
		/* Zap this timer if needed: */
		dev->rxstate_tmo = 0;
		/* If in the MSTX_HOLD state then move back to the MSTX_TXDATA state. */
		if (dev->tx_state == MSTX_HOLD)
			dev->tx_state = MSTX_TXDATA;
	}

	tasklet_schedule(&sltatx_tasklet);
}

/*
 * Function sltatty_write_wakeup (tty)
 *
 *    Called by the driver when there's room for more data.  If we have
 *    more packets to send, we send them here.
 *
 */
static void sltatty_write_wakeup(struct tty_struct *tty)
{
	struct sltatty_cb *priv = tty->disc_data;
	struct slta_dev *dev;

	SLTA_ASSERT(priv != NULL, return;);
	SLTA_ASSERT(priv->magic == SLTATTY_MAGIC, return;);
	dev = priv->dev;

	tty->flags &= ~(1 << TTY_DO_WRITE_WAKEUP);
	tasklet_schedule(&sltatx_tasklet);
}

/* ------------------------------------------------------- */

/*
 * Function sltatty_stop_receiver (tty, stop)
 *
 */

static inline void sltatty_stop_receiver(struct tty_struct *tty, int stop)
{
	struct ktermios old_termios;
	int cflag;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,12, 0)
	down_write(&tty->termios_rwsem);
#else
	mutex_lock(&tty->termios_mutex);
#endif

#if  LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0)
	old_termios = tty->termios;
	cflag = tty->termios.c_cflag;
#else
	old_termios = *(tty->termios);
	cflag = tty->termios->c_cflag;
#endif

	if (stop)
		cflag &= ~CREAD;
	else
		cflag |= CREAD;

#if  LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0)
	tty->termios.c_cflag = cflag;
#else
	tty->termios->c_cflag = cflag;
#endif

	tty->ops->set_termios(tty, &old_termios);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,12, 0)
	up_write(&tty->termios_rwsem);
#else
	mutex_unlock(&tty->termios_mutex);
#endif
}

/*****************************************************************/

/* serialize ldisc open/close with slta_dev */
static DEFINE_MUTEX( sltatty_lock);

/* notifier from slta_dev when slta% device gets opened (ifup) */

static int sltatty_start_dev(struct slta_dev *dev)
{
	struct sltatty_cb *priv;
	struct tty_struct *tty;
	struct tty_driver *drv;

	SLTA_DEBUG(3, "%s()\n", __func__);

	/* serialize with ldisc open/close */
	mutex_lock(&sltatty_lock);

	priv = dev->priv;
	if (unlikely(!priv || priv->magic != SLTATTY_MAGIC))
	{
		mutex_unlock(&sltatty_lock);
		return -ESTALE;
	}

	tty = priv->tty;
	drv = tty->driver;

	start_tty(tty);

	/* Make sure we can receive more data */
	sltatty_stop_receiver(tty, false);

	mutex_unlock(&sltatty_lock);

	/* send initial setup packets, e.g. reset and flush cancel */
	tasklet_schedule(&sltatx_tasklet);

	return 0;
}

#if 0
/* notifier from slta_dev when slta% device gets closed (ifdown) */
static int sltatty_stop_dev(struct slta_dev *dev)
{
	struct sltatty_cb *priv;
	struct tty_struct *tty;

	/* serialize with ldisc open/close */
	mutex_lock(&sltatty_lock);

	priv = dev->priv;
	if (unlikely(!priv || priv->magic != SLTATTY_MAGIC))
	{
		mutex_unlock(&sltatty_lock);
		return -ESTALE;
	}

	tty = priv->tty;

	/* Make sure we don't receive more data */
	sltatty_stop_receiver(tty, true);

	tty->ops->stop(tty);

	mutex_unlock(&sltatty_lock);

	return 0;
}
#endif

/*
 * This is the network interface 'background' service function.
 *
 */
static void slta_transmit_guard(unsigned long arg)
{
	struct slta_dev *dev = (struct slta_dev *) arg;

	SLTA_DEBUG(5, "%s() rx_state=%d tx_state=%d mip_state=%d bpt_count=%d txwait1_tmo=%d\n", __func__, dev->rx_state, dev->tx_state, dev->mip_state, dev->bpt_count, dev->txwait1_tmo);

	slta_tx_process(dev, 0);
}

#if 0
/*
 * Internal flag options for termios setting behavior
 */
#define TERMIOS_FLUSH   1
#define TERMIOS_WAIT    2
#define TERMIOS_TERMIO  4

static void unset_locked_termios(struct termios *termios,
		struct termios *old,
		struct termios *locked)
{
	int i;

#define NOSET_MASK(x, y, z) (x = ((x) & ~(z)) | ((y) & (z)))

	if (!locked)
	{
		pr_warn("Warning?!? termios_locked is NULL.\n");
		return;
	}

	NOSET_MASK(termios->c_iflag, old->c_iflag, locked->c_iflag);
	NOSET_MASK(termios->c_oflag, old->c_oflag, locked->c_oflag);
	NOSET_MASK(termios->c_cflag, old->c_cflag, locked->c_cflag);
	NOSET_MASK(termios->c_lflag, old->c_lflag, locked->c_lflag);
	termios->c_line = locked->c_line ? old->c_line : termios->c_line;
	for (i = 0; i < NCCS; i++)
	termios->c_cc[i] = locked->c_cc[i] ?
	old->c_cc[i] : termios->c_cc[i];
}

static void change_termios(struct tty_struct *tty, struct termios *new_termios)
{
	int canon_change;
	struct termios old_termios = *tty->termios;
	struct tty_ldisc *ld;

	/*
	 *      Perform the actual termios internal changes under lock.
	 */

	/* FIXME: we need to decide on some locking/ordering semantics
	 for the set_termios notification eventually */
	mutex_lock(&tty->termios_sem);

	*tty->termios = *new_termios;
	unset_locked_termios(tty->termios, &old_termios, tty->termios_locked);
	canon_change = (old_termios.c_lflag ^ tty->termios->c_lflag) & ICANON;
	if (canon_change)
	{
		memset(&tty->read_flags, 0, sizeof(tty->read_flags));
		tty->canon_head = tty->read_tail;
		tty->canon_data = 0;
		tty->erasing = 0;
	}

	if (canon_change && !L_ICANON(tty) && tty->read_cnt)
	/* Get characters left over from canonical mode. */
	wake_up_interruptible(&tty->read_wait);

	/* See if packet mode change of state. */

	if (tty->link && tty->link->packet)
	{
		int old_flow = ((old_termios.c_iflag & IXON) &&
				(old_termios.c_cc[VSTOP] == '\023') &&
				(old_termios.c_cc[VSTART] == '\021'));
		int new_flow = (I_IXON(tty) &&
				STOP_CHAR(tty) == '\023' &&
				START_CHAR(tty) == '\021');
		if (old_flow != new_flow)
		{
			tty->ctrl_status &= ~(TIOCPKT_DOSTOP | TIOCPKT_NOSTOP);
			if (new_flow)
			tty->ctrl_status |= TIOCPKT_DOSTOP;
			else
			tty->ctrl_status |= TIOCPKT_NOSTOP;
			wake_up_interruptible(&tty->link->read_wait);
		}
	}

	if (tty->driver->set_termios)
	(*tty->driver->set_termios) (tty, &old_termios);

	ld = tty_ldisc_ref(tty);
	if (ld != NULL)
	{
		if (ld->set_termios)
		(ld->set_termios) (tty, &old_termios);
		tty_ldisc_deref(ld);
	}
	mutex_unlock(&tty->termios_sem);
}

static int set_termios(struct tty_struct *tty, void __user *arg, int opt)
{
	struct termios tmp_termios;
	struct tty_ldisc *ld;
	int retval = tty_check_change(tty);

	if (retval)
	return retval;

	if (opt & TERMIOS_TERMIO)
	{
		memcpy(&tmp_termios, tty->termios, sizeof(struct termios));
		if (user_termio_to_kernel_termios(&tmp_termios,
						(struct termio __user *) arg))
		return -EFAULT;
	}
	else
	{
		if (user_termios_to_kernel_termios(&tmp_termios,
						(struct termios __user *) arg))
		return -EFAULT;
	}

	ld = tty_ldisc_ref(tty);

	if (ld != NULL)
	{
		if ((opt & TERMIOS_FLUSH) && ld->flush_buffer)
		ld->flush_buffer(tty);
		tty_ldisc_deref(ld);
	}

	if (opt & TERMIOS_WAIT)
	{
		tty_wait_until_sent(tty, 0);
		if (signal_pending(current))
		return -EINTR;
	}

	change_termios(tty, &tmp_termios);
	return 0;
}

#endif

/*
 * Function sltatty_ioctl (tty, file, cmd, arg)
 *
 */
static int sltatty_ioctl(struct tty_struct *tty, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct sltatty_info
	{
		char name[12];
	} info;
	struct slta_dev *dev;
	struct sltatty_cb *priv = tty->disc_data;
	int err = 0;
	unsigned int unum;

	SLTA_ASSERT(priv != NULL, return -ENODEV;);
	SLTA_ASSERT(priv->magic == SLTATTY_MAGIC, return -EBADR;);

	SLTA_DEBUG(3, "%s: cmd = 0x%X\n", __func__, cmd);

	dev = priv->dev;
	SLTA_ASSERT(dev != NULL, return -1;);

	switch (cmd)
	{
	case SLTA_IOCGET:
		/*SLTA_ASSERT(dev->netdev != NULL, return -1;); */

		memset(&info, 0, sizeof(info));
		strncpy(info.name, "SLTA-10", sizeof(info.name) - 1);

		if (copy_to_user((void __user *) arg, &info, sizeof(info)))
			err = -EFAULT;
		break;

	case SLTA_IOCBAUD:
		if (copy_from_user(&unum, (void __user *) arg, sizeof(unum)))
		err = -EFAULT;
		else
		sltatty_change_speed(dev, unum);
		break;

	case SLTA_IOCRESET:
		slta_msg_reset(dev);
		/* slta_do_reset(dev); */
		/* tasklet_schedule(&sltatx_tasklet); */
		break;

#if 0
	case TCGETS:
		if (kernel_termios_to_user_termios((struct termios __user *) arg, tty->termios))
			return -EFAULT;
		break;

	case TCSETSF:
		return set_termios(tty, arg, TERMIOS_FLUSH | TERMIOS_WAIT);

	case TCSETSW:
		return set_termios(tty, arg, TERMIOS_WAIT);

	case TCSETS:
		return set_termios(tty, arg, 0);

	case TCSETAF:
		return set_termios(tty, arg, TERMIOS_FLUSH | TERMIOS_WAIT | TERMIOS_TERMIO);

	case TCSETAW:
		return set_termios(tty, arg, TERMIOS_WAIT | TERMIOS_TERMIO);

	case TCSETA:
		return set_termios(tty, arg, TERMIOS_TERMIO);
#endif

	default:
		err = -ENOIOCTLCMD;
		break;
	}
	return err;
}

/*
 *  Function sltatty_open(tty)
 *
 *    This function is called by the TTY module when the Slta line
 *    discipline is called for.  Because we are sure the tty line exists,
 *    we only have to link it to a free Slta channel.
 */
static int sltatty_open(struct tty_struct *tty)
{
	struct slta_dev *dev;
	struct sltatty_cb *priv;
	int i;
	int ret = 0;

	SLTA_DEBUG(3, "%s()\n", __func__);

	/* First make sure we're not already connected. */
	if (tty->disc_data != NULL)
	{
		priv = tty->disc_data;
		if (priv && priv->magic == SLTATTY_MAGIC)
		{
			SLTA_DEBUG(2, "%s(): slta already enabled for this tty\n", __func__);
			ret = -EEXIST;
			goto out;
		}
		tty->disc_data = NULL; /* ### */
	}

	/* stop the underlying  driver */
	sltatty_stop_receiver(tty, true);

	stop_tty(tty);
	if (tty->ops->flush_buffer)
		tty->ops->flush_buffer(tty);

	/* request low-latency delivery */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 9, 0)
	tty->port->low_latency = 1;
#else
	tty->low_latency = 1;
#endif

	/* get a device instance for this driver */
	i = sizeof(*dev);
	dev = kmalloc(i, GFP_KERNEL);
	if (!dev)
	{
		ret = -ENOMEM;
		goto out;
	}
	memset(dev, 0, i);
	SLTA_DEBUG(2, "%s(): dev=%p, %u bytes\n", __func__, dev, i);

	/* allocate private device info block */
	i = sizeof(*priv);
	priv = kmalloc(i, GFP_KERNEL);
	if (!priv)
	{
		kfree(dev);
		ret = -ENOMEM;
		goto out;
	}
	memset(priv, 0, i);
	SLTA_DEBUG(2, "%s(): priv=%p, %u bytes\n", __func__, priv, i);

	priv->magic = SLTATTY_MAGIC;
	priv->tty = tty;
	priv->dev = dev;

	/* Init the various states */
	dev->rx_state = MSRX_IDLE;
	dev->tx_state = MSTX_IDLE;
	dev->mip_state = MIP_IDLE;
	dev->SendRetries = MAX_SEND_RETRIES;
	dev->SendNackRetries = MAX_SEND_NACK_RETRIES;

	/* Build interface buffers out of allocated memory. */
	i = 2 * slta_buffers * sizeof(Ibuffer);
	dev->heap_ptr = kmalloc(i, GFP_KERNEL);
	if (!dev->heap_ptr)
	{
		kfree(dev);
		kfree(priv);
		ret = -ENOMEM;
		goto out;
	}
	memset(dev->heap_ptr, 0, i);
	SLTA_DEBUG(2, "%s(): heap=%p, %u bytes\n", __func__, dev->heap_ptr, i);

	/* set up a 'background' time service. */
#if LINUX_VERSION_CODE < KERNEL_VERSION(4,14,0)
	init_timer(&dev->slta_timer);
	dev->slta_timer.function = slta_timer_service;
	dev->slta_timer.data = (unsigned long) dev;
#else
	timer_setup(&dev->slta_timer, slta_timer_service, 0);
	/* the third argument may include TIMER_* flags */
#endif
	dev->slta_timer.expires = jiffies + msecs_to_jiffies(TIMER_INTERVALL);

	/* adjust the tasklet data fields */
	sltatx_tasklet.data = (unsigned long) dev;

	/* receive wait queue */
	init_waitqueue_head(&dev->wr_wait);
	init_waitqueue_head(&dev->rd_wait);
	spin_lock_init(&dev->sxo_lock);
	spin_lock_init(&dev->sxi_lock);

	/* serialize with start_dev - in case we were racing with ifup */
	mutex_lock(&sltatty_lock);

	dev->priv = priv;
	tty->disc_data = priv;
	tty->receive_room = 65536;

	/* init buffer queues */
	slta_msg_reset(dev);

	add_timer(&dev->slta_timer);

	mutex_unlock(&sltatty_lock);

	/* callback normally invoked from higher protocol layer */
	/* (for now, we do not have such a layer ...) */
	sltatty_start_dev(dev);

	SLTA_DEBUG(0, "%s - %s: slta line discipline opened\n", __func__,
			tty->name);

	/* Get the SLTA-10 out of the unconfigured state */
	/* slta_do_reset(dev); */
	/* slta_do_flush_cancel(dev); */

	return 0;
	out:
	return ret;
}

/*
 *  Function sltatty_close (tty)
 *
 *    Close down a Slta channel. This means flushing out any pending queues,
 *    and then restoring the TTY line discipline to what it was before it got
 *    hooked to Slta (which usually is TTY again).
 */
static void sltatty_close(struct tty_struct *tty)
{
	struct sltatty_cb *priv = tty->disc_data;
	struct slta_dev *dev = priv->dev;

	SLTA_ASSERT(priv != NULL, return;);
	SLTA_ASSERT(priv->magic == SLTATTY_MAGIC, return;);

	/* we are dead now */
	tty->disc_data = NULL;

	/* Stop timer service */
	del_timer_sync(&dev->slta_timer);

	/* Stop tty */
	sltatty_stop_receiver(tty, true);
	tty->flags &= ~(1 << TTY_DO_WRITE_WAKEUP);

	stop_tty(tty);
	if (tty->ops->flush_buffer)
		tty->ops->flush_buffer(tty);

	/* release resources */
	kfree(priv);
	kfree(dev->heap_ptr);
	kfree(dev);

	SLTA_DEBUG(0, "%s - %s: slta line discipline closed\n", __func__,
			tty->name);
}

/*
 * This is the network interface read function. If a message is enqueued
 * onto the input buffer linked list then it is transfered into the
 * caller's structure.
 */
static ssize_t sltatty_read(struct tty_struct *tty, struct file *file, unsigned char __user *buf, size_t nr)
{
	struct sltatty_cb *priv = tty->disc_data;
	struct slta_dev *dev;
	unsigned char __user *b = buf;
	unsigned char lbuf[MAX_NI_DATA + 2];
	int rval = 0;

	SLTA_ASSERT(priv != NULL, return -ENODEV;);
	SLTA_ASSERT(priv->magic == SLTATTY_MAGIC, return -EBADR;);
	dev = priv->dev;

	/* verify user access to buffer */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0))
	if (!access_ok(buf, nr))
#else
	if (!access_ok(VERIFY_WRITE, buf, nr))
#endif
	{
		printk(KERN_WARNING "%s: can't verify user buffer\n", __func__);
		return -EFAULT;
	}

	mutex_lock(&sltatty_lock);

	SLTA_DEBUG(2, "%s: ready blocks %d, next block busy %d\n", __func__,
			slta_rx_ready, dev->sxi_tail->busy);

	if (!dev->sxi_tail->busy)
	{

		if (file->f_flags & O_NONBLOCK)
		{
			mutex_unlock(&sltatty_lock);
			return -EAGAIN;
		}

		SLTA_DEBUG(2, "%s: read is going to sleep\n", __func__);

		if (wait_event_interruptible(dev->rd_wait,
						(dev->sxi_tail->busy)))
		{
			mutex_unlock(&sltatty_lock);
			return -ERESTARTSYS;
		}
		SLTA_DEBUG(2, "%s: blocking read woke up\n", __func__);
	}

	spin_lock_irq(&dev->sxi_lock);

	SLTA_DEBUG(14, "%s: deliver packet @%p\n", __func__, dev->sxi_tail);

	/* Copy to local buffer */
	lbuf[0] = dev->sxi_tail->llh.ni_cmd;
	if ((dev->sxi_tail->llh.length) > 0)
		lbuf[1] = dev->sxi_tail->llh.length - 1;
	else
		lbuf[1] = 0;
	memcpy(lbuf + 2, dev->sxi_tail->llm.ni_data, lbuf[1]);
	rval = lbuf[1] + 2;

	/* Check for queued errors: */
	if (dev->sxi_tail->error)
	{
		SLTA_DEBUG(1, "%s(): buffer has error flag set\n", __func__);
		/* rval = -EIO; */
	}

	/* free this buffer */
	dev->sxi_tail->error = 0;
	dev->sxi_tail->busy = false;
	dev->sxi_tail = dev->sxi_tail->next;

	/* recalculate free receiver buffer space */
	slta_rx_ready--;
	sltatty_set_room(tty);

	spin_unlock_irq(&dev->sxi_lock);

	if (slta_debug > 2)
	DumpBuffer("<r", (void *) lbuf, lbuf[1]);

	/* Transfer to userspace */
	if (copy_to_user(b, lbuf, lbuf[1] + 2))
	rval = -EFAULT;

	mutex_unlock(&sltatty_lock);

	return rval;
}

/*
 * This is the network interface write function. The message is enqueued onto
 * the output buffer linked list, followed by a call to ldv_post_events().
 * Return LDV_NO_BUFF_AVAIL if there are no output buffers available at this
 * layer.
 */
static ssize_t sltatty_write(struct tty_struct *tty, struct file *file,
		const unsigned char *buf, size_t nr)
{
	struct sltatty_cb *priv = tty->disc_data;
	struct slta_dev *dev;
	int rval = nr;

	SLTA_ASSERT(priv != NULL, return -ENODEV;);
	SLTA_ASSERT(priv->magic == SLTATTY_MAGIC, return -EBADR;);
	dev = priv->dev;

	if (nr > MAX_NI_DATA)
		return -EIO;

	mutex_lock(&sltatty_lock);

	if (dev->sxo_head->busy)
	{

		if (file->f_flags & O_NONBLOCK)
		{
			mutex_unlock(&sltatty_lock);
			return -EAGAIN;
		}

		if (wait_event_interruptible(dev->wr_wait,
				(!dev->sxo_head->busy)))
		{
			mutex_unlock(&sltatty_lock);
			return -ERESTARTSYS;
		}

		SLTA_DEBUG(2, "%s: blocking write woke up\n", __func__);
	}

	if (slta_debug > 2)
		DumpBuffer(">w", (void *) buf, nr);

	spin_lock_irq(&dev->sxo_lock);

	memcpy(dev->sxo_head->llm.ni_data, buf + 2, nr - 2);
	dev->sxo_head->llh.ni_cmd = buf[0];
	dev->sxo_head->llh.length = buf[1];

	SLTA_DEBUG(3, "%s: cmd = 0x%X, len = %u bytes\n", __func__,
			dev->sxo_head->llh.ni_cmd, dev->sxo_head->llh.length);

	/* Bump length to include command_queue */
	dev->sxo_head->llh.length++;

	/* Fire! */
	dev->sxo_head->busy = true;
	dev->sxo_head = dev->sxo_head->next;

	spin_unlock_irq(&dev->sxo_lock);

	/* schedule transmit process, if not already active */
	tasklet_schedule(&sltatx_tasklet);

#if 0
	if (dev->g_tx_error)
	{
		dev->g_tx_error = false;
		rval = -EIO;
	}
#endif

	mutex_unlock(&sltatty_lock);

	return rval;
}

/**
 *	sltatty_poll		-	poll method for N_SLTA
 *	@tty: terminal device
 *	@file: file accessing it
 *	@wait: poll table
 *
 *	Called when the line discipline is asked to poll() for data or
 *	for special events. This code is not serialized with respect to
 *	other events save open/close.
 *
 *	This code must be sure never to sleep through a hangup.
 *	Called without the kernel lock held - fine
 *
 *	FIXME: if someone changes the VMIN or discipline settings for the
 *	terminal while another process is in poll() the poll does not
 *	recompute the new limits. Possibly set_termios should issue
 *	a read wakeup to fix this bug.
 */

static unsigned int sltatty_poll(struct tty_struct *tty, struct file *file, poll_table *wait)
{
	struct sltatty_cb *priv = tty->disc_data;
	struct slta_dev *dev;
	unsigned int mask = 0;

	SLTA_ASSERT(priv != NULL, return POLLERR;);
	SLTA_ASSERT(priv->magic == SLTATTY_MAGIC, return POLLERR | POLLHUP;);
	dev = priv->dev;

	poll_wait(file, &dev->rd_wait, wait);
	poll_wait(file, &dev->wr_wait, wait);

	spin_lock_irq(&dev->sxi_lock);
	if (slta_rx_ready > 0)
		mask |= POLLPRI | POLLIN | POLLRDNORM;
	spin_unlock_irq(&dev->sxi_lock);

	spin_lock_irq(&dev->sxo_lock);
	if (!dev->sxo_head->busy)
		mask |= POLLOUT | POLLWRNORM;
	spin_unlock_irq(&dev->sxo_lock);

	return mask;
}

/* ------------------------------------------------------- */

static struct tty_ldisc_ops slta_ldisc =
		{
				.magic = TTY_LDISC_MAGIC,
				.name = "slta",
				.open = sltatty_open,
				.close = sltatty_close,
				.read = sltatty_read,
				.write = sltatty_write,
				.ioctl = sltatty_ioctl,
				.poll = sltatty_poll,
				.receive_buf = sltatty_receive_buf,
				.write_wakeup = sltatty_write_wakeup,
				.owner = THIS_MODULE,
		};

/* ------------------------------------------------------- */

static int __init module_sltatty_init(void)
{
	int err;

	SLTA_DEBUG(0, "slta-tty %s: Alert-Ack=%d  reliable transport protocol=%d\n", DRIVER_VERSION, slta_alert_ack, USE_RELIABLE);

	err = tty_register_ldisc(N_SLTA, &slta_ldisc);
	if (err)
		SLTA_ERROR("SLTA: can't register line discipline (err = %d)\n", err);
	return err;
}

static void __exit module_sltatty_cleanup(void)
{
	int ret;

	ret = tty_unregister_ldisc(N_SLTA);
	if (ret)
		SLTA_ERROR("%s(), can't unregister line discipline (%d)\n", __func__, ret);
}
module_init( module_sltatty_init);
module_exit( module_sltatty_cleanup);

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
