/*
 *  (based on original Echelon Solaris driver LDVSLTA)
 */


#define DRIVER_VERSION  "V2.06"
#define DRIVER_AUTHOR   "Michael Tross <mtr@nets-solutions.de>, Volker Schober <vschober@gesytec.de>"
#define DRIVER_DESC     "Serial LonTalk Adapter SLTA device driver"

#if defined(__KERNEL__)

#define USE_RELIABLE 01

// Base timing intercall
#define TIMER_INTERVALL 10

// Some timing constants
#define ACK_WAIT_PD		(100  / (TIMER_INTERVALL))	// ms
#define RXWAIT1_TMO		(100  / (TIMER_INTERVALL))	// ms
#define TXWAIT1_TMO		(100  / (TIMER_INTERVALL))	// 100 ms alert/ack timeout
#define TXWAIT2_TMO		(1000 / (TIMER_INTERVALL))	// 1 sec ack/nack timeout
#define SEND_REPEAT_DELAY       (1000 / (TIMER_INTERVALL))	// 1 sec
#define SEND_NACK_REPEAT_DELAY  (1000 / (TIMER_INTERVALL))	// 1 sec

#define BRQ_RETRIES		50
#define MAX_SEND_RETRIES	5
#define MAX_SEND_NACK_RETRIES	100

#define USBSERIAL_TX_DONE_DELAY	60

#endif // defined(__KERNEL__)

/*
 * line discipline index
 *   see include/asm/termios.h
 * Unfortunately slta is still not part of the main kernel.
 * As of 2.6.16 there are 16 line disciplines registered,
 * so choose a unused number, N_MASC is "Reserved for Mobitex module",
 * but there is no driver using it.
 */
#ifndef N_SLTA
#define N_SLTA N_MASC
#endif

#if defined(__KERNEL__)

/*
 * Some debugging helpers
 */
#define CONFIG_SLTA_DEBUG
#ifdef CONFIG_SLTA_DEBUG

/* use 0 for production, 1 for verification, >2 for debug */
//#define SLTA_DEBUG_LEVEL 2

#define SLTA_DEBUG(n, args...)				\
	do {						\
		if (slta_debug >= (n))			\
			printk(KERN_DEBUG args);	\
	} while (0)

#define SLTA_ASSERT(expr, func)							\
	do {									\
		if (!(expr)) {							\
			printk( "Assertion failed! %s:%s:%d %s\n",		\
					__FILE__,__FUNCTION__,__LINE__,(#expr) );\
			func							\
		}								\
	} while (0)
#define SLTA_ASSERT_LABEL(label)        label

#else

#define SLTA_DEBUG(n, args...) do { } while (0)
#define SLTA_ASSERT(expr, func) do { (void)(expr); } while (0)
#define SLTA_ASSERT_LABEL(label)
#endif /* CONFIG_SLTA_DEBUG */

#define SLTA_WARNING(args...) do { printk(KERN_WARNING args); } while (0)
#define SLTA_MESSAGE(args...) do { printk(KERN_INFO args); } while (0)
#define SLTA_ERROR(args...)   do { printk(KERN_ERR args); } while (0)

/*
 * Pretty-Print IO buffer
 */
static inline void DumpBuffer(char *prefix, void *pvBuffer, unsigned int length)
{
	pr_notice("%s\n", prefix);
	print_hex_dump(KERN_CONT, "", DUMP_PREFIX_OFFSET,
			16, 1,
			pvBuffer, length, false);
}

/*
 *  Random numbers which must be unique to give the best protection
 */
#define SLTATTY_MAGIC        0x422907

/* define the max network interface transfer size. */
#define MAX_NI_DATA		256

/* Receiver states: */
typedef enum {
    MSRX_IDLE,
    MSRX_CLASH,
    MSRX_RXHEAD,
    MSRX_RXDATA
} MsrxState;

/* Transmitter states: */
typedef enum {
    MSTX_IDLE,
    MSTX_WAIT1,		// 1    waiting for NI_ALERT_ACK
    MSTX_HOLD,		// 2    holding off downlink transfer
    MSTX_TXHEAD,    // 3
    MSTX_TXDATA,    // 4
    MSTX_TXCS,      // 5
    MSTX_TXDONE		// 6    when 'reliable' protocol.
} MstxState;

/* Buffer request states: */
typedef enum {
    MIP_IDLE,
    MIP_ACK_WAIT,
    MIP_ACK_NACK,
    MIP_ACK_OKAY
} MipState;

#define NI_ALERT		0x01
#define NI_ALERT_ACK	0xFE
#define RM_ACK		    0x06	// Standard ASCII ACK
#define RM_NACK		    0x15	// Standard ASCII NACK

#define niTYPE			0xF0	/* type mask */

/* Queue Types */
#define niLOCAL_CMD		0x00	/* local command */
#define niTQ			0x02	/* trans queue */
#define niTQ_P			0x03	/* trans queue priority */
#define niNTQ			0x04	/* non-trans queue */
#define niNTQ_P			0x05	/* non-trans queue priority */
#define niRESPONSE		0x06	/* response queue */
#define niINCOMING		0x08	/* incoming queue */

#define niCOMM			0x10	/* outgoing message */
#define niNETMGMT		0x20	/* local net mgmt */

/* Command/queue byte values: */
#define niRESET			0x50
#define niFLUSH_CANCEL	0x60
#define niACK			0xC0
#define niNACK			0xC1
#define niPUPXOFF		0xE1
#define niPUPXON		0xE4
#define niSSTATUS		0xE0	/* u+d, SLTA only */

/* The structure for the SLTA link-layer header: */
typedef struct Llhed_s {
	u8 alert;			/* always MI_ALERT		*/
#if USE_RELIABLE
	u8 seqn;          // sequence number
#endif
	u8 length;		/* includes ni_cmd, excludes checksum.	*/
	u8 nlength;		/* ones compl. of length	*/
	u8 ni_cmd;		/* the command/queue byte	*/
} Llhed;

/* The structure for the SLTA link-layer message data: */
typedef struct Llmsg_s {
	u8 ni_data[MAX_NI_DATA+1];
} Llmsg;

/* The structure for the input and output interface buffer linked lists: */
typedef struct Ibuffer_s {
	struct Ibuffer_s *next;
	u8 busy;
	u8 error;
	Llhed llh;
	Llmsg llm;
} Ibuffer;

/* The structure for the output buffer head: if llh.length > 1 then
** a message array will be sent after the llh struct.			*/
typedef struct {
	Llhed llh;
	u8 checksum;
} Obuffer;

typedef struct {
	u8 command_queue;
	u8 length;		/* size of .message[] */
	u8 message[MAX_NI_DATA+1];
} AppBuffer;

typedef struct {
	int brqs_sent;
	int mbfs_sent;
	int bracks_rcvd;
	int brnacks_rcvd;
	int alacks_rcvd;
	int clashes;
} DbgData;

/* Network Driver Return Codes */
typedef enum {
	LDV_OK			= 0,
	LDV_NOT_FOUND		= 1,
	LDV_ALREADY_OPEN	= 2,
	LDV_DEVICE_ERR		= 4,
	LDV_NO_MSG_AVAIL	= 6,
	LDV_NO_BUFF_AVAIL	= 7
} LDVCode;

struct slta_dev {
	char hwname[32];

	/* System & kernel ressources */
	atomic_t enable_rx;
	spinlock_t sxo_lock, sxi_lock;
	struct timer_list slta_timer;
	wait_queue_head_t rd_wait, wr_wait;
	unsigned long flags;

	/* Options */
	int			raw_tx;
	int			new_speed;
	unsigned	speed;

	/* Device operating data */
	MsrxState rx_state;		/* The receiver state		*/
	MstxState tx_state;		/* The transmitter state	*/
	MipState mip_state;		/* buffer request state		*/
	u8 *bpt;			/* tx pointer			*/
	int bpt_count;			/* tx count - works with tx_state. */

	bool g_tx_error;		/* downlink error flag	*/

	/* Timers: */
	int ack_wait_timer;
	int rxstate_tmo;
	int txwait1_tmo;

	/* Input buffer list */
	Ibuffer *sxi_head, *sxi_tail;
	u8 *rxb_ptr;			/* pointer to current receive buffer */
	int rxb_indx;			/* keeps track of receive size */
	int rxb_drop;			/* signal to drop incoming data */

	/* Ouput buffer list */
	Ibuffer *sxo_head, *sxo_tail;
	void *heap_ptr;

	/* These form local link layer header & message pointers */
	Llhed dl_head;
	u8 mstx_chksum;

	bool rx_try_flag;

	u8 transmit_sequence_n;
	u8 receive_sequence_n;

	int     SendRetries, SendNackRetries;

	DbgData dbg;

	void *priv;
};

struct sltatty_cb {
	u32 magic;

	struct slta_dev *dev;
	struct tty_struct  *tty;
};

#endif /* __KERNEL */

#define SLTA_IOC_MAGIC 's'
#define SLTA_IOCGET     _IOR(SLTA_IOC_MAGIC, 2, struct sltatty_info)
#define SLTA_IOCBAUD    _IOW(SLTA_IOC_MAGIC, 3, unsigned int)
#define SLTA_IOCRESET   _IO(SLTA_IOC_MAGIC, 4)
#define SLTA_IOC_MAXNR   4
